#
#                                                                      #
# ==================================================================== #
# IMPORT built-in modules                                              #
# ==================================================================== #
#                                                                      #
import os
import os.path
import sys
from copy import deepcopy
#                                                                      #
# ==================================================================== #
# IMPORT modules made on purpose                                       #
# ==================================================================== #
#                                                                      #
#
# get path to generic_frame.fluka
from find_paths import GENERIC_FRAME
print(GENERIC_FRAME)
#
loc_path = os.path.realpath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.insert(0,loc_path)
from structure import present_FEDB
#
from FEDB_ELEMENTs import parse_prototypes_lbp, synch_against_BASIC_ELEMENTs, \
                          checkFlukaNameUniqueness, REPLICA
#
from COMPLEX_STRING import error_message, secure_open_file_4_w, red, grn, \
                           secure_open_file_4_r, secure_open_file_4_a
#
from FLUKA_lib import INCLUDE_STATEMENT_FMT
#
from LB_sty import HEADERs

#                                                                      #
# ==================================================================== #
#                               FUNCTIONs                              #
# ==================================================================== #
#                                                                      #

def start():

    print("")
    print(" " + sys.argv[0].split( "/" )[-1] + " <fedb_series> <fedb_tag> <ass_name>")
    print(" by A. Mereghetti")
    print(" ")
    print(" for creating the .inp file to test an assembly")
    print(" NB: if <ass_name> is not given, then <fedb_tag> will be used")
    print("")
    if ( len( sys.argv ) < 3 ):
        msg = "Please specify:\n"
        msg = msg + "a  <fedb_series> (mandatory); \n"
        msg = msg + "a  <fedb_tag>    (mandatory); \n"
        msg = msg + "an <ass_name>    (optional)."
        error_message( msg, True )

    # get input information:
    fedb_series = sys.argv[1]
    fedb_tag    = sys.argv[2]
    if ( len( sys.argv ) == 3 ):
        ass_name = fedb_tag[:]
    else:
        ass_name = sys.argv[3]

    ass_file_name = present_FEDB.inflect( "name", fedb_series, fedb_tag )
    ass_full_path = present_FEDB.inflect( "smbl", fedb_series, fedb_tag )

    # output files:
    out_file_name  = present_FEDB.inflect( "name", fedb_series, ass_name ) + ".inp"
    log_file_name  = present_FEDB.inflect( "name", fedb_series, ass_name ) + ".log"

    # initialise log file:
    logfile = secure_open_file_4_w( log_file_name )
    logfile.write( " called from:\n %s \n" % ( sys.argv[0] ) )
    logfile.close()

    return out_file_name, log_file_name, ass_full_path, ass_name

#                                                                      #
# ==================================================================== #
#   MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN MAIN   #
# ==================================================================== #
#                                                                      #

# some initialisation
out_file_name, log_file_name, ass_full_path, ass_name = start()
position = [ 0.0, 0.0, 0.0 ]
eulers   = [ 0.0, 0.0, 0.0 ] # [deg]
ADDITIONAL_TRANSF = []
#
l_Bl = 2 # Max number of beams
# A.Mereghetti, 09/08/2012
# with devel, this dictonary should be replaced by a more complex
#      dictionary in customize_geo_builder.py
flukaIndices = { "ilatt": 0, "irot": 0 }
#
# exceptions in the mapping of prototypes:
# these are arrays of Dictionaries, with a Dictionary per beam;
# each dictionary stores the name of an element as key,
#   and the name of the corresponding ELEMENT_MODEL as value;
# eg:
#   mappingExceptions["map_entries"][0]={
#                "MCBY.A4R2":"MCBYrob",
#                "MCBY.B4R2":"MCBYang",
#                "MCBY.C4R2":"MCBYtri",
#                }
mappingExceptions={ \
    "map_entries" : [ {} for col in range( l_Bl ) ] ,\
    "map_vetos"   : [ {} for col in range( l_Bl ) ]  \
        }

# log FEDB
logfile = secure_open_file_4_a( log_file_name )
logfile.write( present_FEDB.log_structure() )
logfile.close()

# parse the assembly file, and return the BASIC_ELEMENTs and the
#     ELEMENT_MODELs
BASIC_ELEMENTs, ELEMENT_MODELs, mappingExceptions = \
    parse_prototypes_lbp( ass_full_path, present_FEDB, \
                              mappingExceptions, log_file_name )
if ( ass_name not in ELEMENT_MODELs ):
    msg = "No '%s' ASSEMBLY declared in file: \n" % ( red( ass_name ) )
    msg = msg + "'%s'" % ( red( ass_full_path ) )
    error_message( msg, True )

# log BASIC_ELEMENTs:
logfile = secure_open_file_4_a( log_file_name )
for element in BASIC_ELEMENTs.values():
    logfile.write( element.log() )
logfile.close()

# synchronise the ELEMENT_MODELs against the BASIC_ELEMENTs
ELEMENT_MODELs = synch_against_BASIC_ELEMENTs( ELEMENT_MODELs, BASIC_ELEMENTs, \
                                                   present_FEDB, log_file_name )

# build internal rotdefis, in order to set up the assembly
for ele_name, model in ELEMENT_MODELs.items():
    model.build_rotdefis()
    continue

# check uniqueness of FLUKA names:
checkFlukaNameUniqueness( ELEMENT_MODELs, log_file_name )

# log ELEMENT_MODELs:
logfile = secure_open_file_4_a( log_file_name )
for element in ELEMENT_MODELs.values():
    logfile.write( element.log() )
logfile.close()

# take the only ELEMENT_MODEL present in the dictionary:
current_replica = REPLICA()
current_replica.inherit_from( ELEMENT_MODELs[ ass_name ] )
current_replica.names["full"] = ass_name
current_replica.names["prto"] = ass_name
current_replica.names["FLKA"] = ass_name
current_replica.build_rotdefis( position, eulers, "", ADDITIONAL_TRANSF )
# A.Mereghetti, 28/06/2012
# rename only the transformations, in order to avoid
#     names longer than 10 characters (geoedit complains);
# bodies and regions will keep the names found in the .lbp assembly file,
#     so that geometry debugging becomes easier;
# A.Mereghetti, 09/08/2012
# starting from devel, these two calls should be changed into:
#    current_replica.update_names_and_indeces( key="TRANSFs" )
current_replica.change_names( key="TRANSFs" )
current_replica.change_indices( key="TRANSFs", flukaIndices=flukaIndices )

BODIEs  = current_replica.ret( "BODIEs"  )
REGIONs = current_replica.ret( "REGIONs" )
TRANSFs = current_replica.ret( "TRANSFs" )

# vacuum around assembly:
EXTERNAL_REG = deepcopy( current_replica.ret( "EXTERNAL" ) )
EXTERNAL_REG.name     = "VAROUND"
EXTERNAL_REG.material = "VACUUM"
EXTERNAL_REG.comment = "* vacuum around assembly"
zones = EXTERNAL_REG.defin.split( "|" )
EXTERNAL_REG.defin = ""
if ( len( zones ) > 1 ):
    for zone in zones:
        EXTERNAL_REG.defin = EXTERNAL_REG.defin + " | +%-8s %s" % ( "cont", zone.strip() )
        continue
else:
    EXTERNAL_REG.defin = " +%-8s %s" % ( "cont", zones[0].strip() )
REGIONs.append( EXTERNAL_REG )

# write everything
print("")
print(" dumping the geometry in the '%s' file..." % \
    ( grn( out_file_name ) ))
ofile = secure_open_file_4_w( out_file_name )
ifile = secure_open_file_4_r( GENERIC_FRAME )
flag = 0
fastforward_def = 'the_big_bang_theory_is_a_silly_sit_com'
fastforward = fastforward_def
#
for line in ifile.readlines():

    #
    # understand what to do:
    # NB: multi-line actions
    #

    if ( line.find("$START:build_line:BODIEs$") > -1 ):
        ofile.write( line )
        flag = 1
    elif ( line.find("$START:build_line:REGIONs$") > -1 ):
        ofile.write( line )
        flag = 2
    elif ( line.find("$START:build_line:LATTICEs$") > -1 ):
        ofile.write( line )
        flag = 3
    elif ( line.find("$START:build_line:ROT-DEFIs$") > -1 ):
        ofile.write( line )
        flag = 4
    elif ( line.find("$START:build_line:ASSIGNMAs$") > -1 ):
        ofile.write( line )
        flag = 6
    elif ( line.find("$START:build_line:PARKING_region$") > -1 ):
        ofile.write( line )
        flag = 8
    elif ( line.find( "MGNFIELD" ) > -1 ):
        # don't copy MGNFIELD card
        continue
    elif ( line.find( fastforward ) > -1 ):
        ofile.write( line )
        fastforward = fastforward_def
        flag = 0
        continue

    #
    # ...and do it!
    #

    # flag == -1: skip line, and do not check the other
    #   values of flag
    if ( flag == -1 ):
        continue

    # flag == 0: copy line
    elif ( flag == 0 ):
        ofile.write( line )

    # flag == 1: bodies declarations
    elif ( flag == 1 ):

        print(" ...dumping FLUKA BODIEs;")
        ofile.write( "* \n" )

        # BASIC ELEMENTs:
        for basic_element in BASIC_ELEMENTs.values():
            line = basic_element.ret_FLUKA_include_statement( "body" )
            ofile.write( "%s \n" % ( line ) )
            ofile.write( "* \n" )

        # a header:
        ofile.write( HEADERs["assembly"] % ( "bodies for assembly" ) + "\n" )

        # BODIEs:
        for body in BODIEs:
            body.convert2infinite()
            ofile.write( body.returnSTRING() + "\n" )
            ofile.write( "* \n" )

        ofile.write( "* \n" )
        fastforward = "$END:build_line:BODIEs$"
        flag = -1

    # flag == 2: regions declarations
    elif ( flag == 2 ):

        print(" ...dumping FLUKA REGIONs;")
        ofile.write( "* \n" )

        # BASIC elements:
        for basic_element in BASIC_ELEMENTs.values():
            line = basic_element.ret_FLUKA_include_statement( "regn" )
            ofile.write( "%s \n" % ( line ) )
            ofile.write( "* \n" )

        # a header:
        ofile.write( HEADERs["assembly"] % ( "regions for assembly" ) + "\n" )

        # REGIONs:
        for region in REGIONs:
            ofile.write( "%s \n" % ( region.returnSTRING() ) )
            ofile.write( "* \n" )

        ofile.write( "* \n" )
        fastforward = "$END:build_line:REGIONs$"
        flag = -1

    # flag == 3: LATTICE declarations
    elif ( flag == 3 ):

        print(" ...dumping FLUKA LATTICEs;")
        ofile.write( "* \n" )

        for region in REGIONs:
            if ( region.ilatt > 0 ):
                ofile.write( "%s \n" % ( region.returnLATTICE() ) )

        ofile.write( "* \n" )
        fastforward = "$END:build_line:LATTICEs$"
        flag = -1

    # flag == 4: ROT-DEFI declarations
    elif ( flag == 4 ):

        print(" ...dumping FLUKA ROT-DEFIs;")
        ofile.write( "* \n" )

        for transformation in TRANSFs.values():
            for rotdefi in transformation:
                ofile.write( "%s \n" % ( rotdefi.returnSTRING( insert_comment=True ) ) )

        ofile.write( "* \n" )
        fastforward = "$END:build_line:ROT-DEFIs$"
        flag = -1

    # flag == 6: material assignment
    elif ( flag == 6 ):

        print(" ...dumping FLUKA ASSIGNMA cards;")
        ofile.write( "* \n" )

        # FEDB elements:
        # - materials.inp file:
        for tmpMatFile in present_FEDB.ret_material_files():
            line = INCLUDE_STATEMENT_FMT % ( tmpMatFile )
            ofile.write( "%s \n" % ( line ) )
            ofile.write( "* \n" )
        # - element declarations:
        for basic_element in BASIC_ELEMENTs.values():
            line = basic_element.ret_FLUKA_include_statement( "assm" )
            ofile.write( "%s \n" % ( line ) )
            ofile.write( "* \n" )

        # a header:
        ofile.write( HEADERs["assembly"] % ( "material assignment for assembly" ) + "\n" )

        # REGIONs:
        for region in REGIONs:
            ofile.write( "%s \n" % ( region.returnASSIGNMA() ) )

        ofile.write( "* \n" )
        fastforward = "$END:build_line:ASSIGNMAs$"
        flag = -1

    # flag == 8: new prototypes to be subtracted from the parking:
    elif ( flag == 8 ):

        # FEDB lines:

        if ( len( BASIC_ELEMENTs ) > 0 ):

            print(" ...subtracting new prototypes from parking region;")
            ofile.write( "* \n" )

            for basic_element in BASIC_ELEMENTs.values():
                line = basic_element.ret_FLUKA_subtract_bounding_box_from_parking()
                ofile.write( "%s \n" % ( line ) )

            ofile.write( "* \n" )

        fastforward = "$END:build_line:PARKING_region$"
        flag = -1

    continue

ofile.close()
ifile.close()




# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import numpy as np
from collections.abc import Mapping, Iterator
from dataclasses import is_dataclass, asdict


def _arrays_equal(a, b, *, rtol=1e-7, atol=0.0, debug=False):
    try:
        a = np.asarray(a)
    except:
        if debug: print(f"Failed to convert first argument to array:\n{a}")
        return False
    try:
        b = np.asarray(b)
    except:
        if debug: print(f"Failed to convert second argument to array:\n{b}")
        return False

    if a.shape != b.shape:
        if debug: print(f"Array shapes differ:\n{a}\nvs\n{b}")
        return False

    a_floaty = np.issubdtype(a.dtype, np.floating) or np.issubdtype(a.dtype, np.complexfloating)
    b_floaty = np.issubdtype(b.dtype, np.floating) or np.issubdtype(b.dtype, np.complexfloating)

    if a_floaty or b_floaty:
        # allclose can't handle object dtype; fall back to exact for object arrays
        if a.dtype == object or b.dtype == object:
            eq = np.array_equal(a, b)
            if debug and not eq: print(f"Failed exact object arrays comparison:\n{a}\nvs\n{b}")
            return eq
        eq = np.allclose(a, b, rtol=rtol, atol=atol, equal_nan=True)
        if debug and not eq: print(f"Failed allclose comparison:\n{a}\nvs\n{b}")
        return eq

    return np.array_equal(a, b)


def _is_sequence_like(obj):
    # "sequence-like" is safer than "iterable":
    # it suggests stable ordering + finite length + repeatable access
    if isinstance(obj, (str, bytes, bytearray)):
        return False
    if isinstance(obj, (np.ndarray, Mapping)):
        return False
    return hasattr(obj, "__len__") and hasattr(obj, "__getitem__")


def deep_equal(x, y, *, rtol=1e-7, atol=0.0, expand_numpy_and_hybridclass=False,
               compare_iterators=False, max_depth=200, debug=False):
    def _eq(a, b, depth):
        if a is b:
            return True

        if depth > max_depth:
            raise RecursionError(f"Max depth {max_depth} exceeded (possible cycle).")

        if expand_numpy_and_hybridclass:
            if hasattr(a, 'to_dict') and callable(a.to_dict):
                a = a.to_dict()
            if hasattr(b, 'to_dict') and callable(b.to_dict):
                b = b.to_dict()
            if hasattr(a, 'tolist') and callable(a.tolist):
                a = a.tolist()
            if hasattr(b, 'tolist') and callable(b.tolist):
                b = b.tolist()

        # Cheap type mismatch early
        ta, tb = type(a), type(b)
        if ta is not tb:
            # allow numeric-ish comparisons later; don't immediately fail for numpy scalars etc.
            pass

        # NumPy scalars: compare as Python scalars (fast)
        if isinstance(a, np.generic) or isinstance(b, np.generic):
            if not (isinstance(a, np.generic) and isinstance(b, np.generic)):
                if debug: print(f"Type mismatch at depth {depth} ({type(a)} vs {type(b)}):\n{a}\nvs\n{b}")
                return False
            # float-like uses isclose
            if isinstance(a, (np.floating, np.complexfloating)) or isinstance(b, (np.floating, np.complexfloating)):
                eq = np.isclose(a, b, rtol=rtol, atol=atol, equal_nan=True)
                if debug and not eq: print(f"Failed isclose comparison:\n{a}\nvs\n{b}")
                return eq
            return a == b

        # NumPy arrays / array-like
        if isinstance(a, np.ndarray) or isinstance(b, np.ndarray):
            if not (isinstance(a, np.ndarray) and isinstance(b, np.ndarray)):
                if debug: print(f"Type mismatch at depth {depth} ({type(a)} vs {type(b)}):\n{a}\nvs\n{b}")
                return False
            return _arrays_equal(a, b, rtol=rtol, atol=atol, debug=debug)

        # to_dict only for non-basic objects (avoid repeated hasattr on primitives)
        if not expand_numpy_and_hybridclass:
            if not isinstance(a, (str, bytes, bytearray, int, float, bool, type(None), tuple, list, dict)):
                td = getattr(a, "to_dict", None)
                if td is not None and callable(td):
                    a = td()
            if not isinstance(b, (str, bytes, bytearray, int, float, bool, type(None), tuple, list, dict)):
                td = getattr(b, "to_dict", None)
                if td is not None and callable(td):
                    b = td()

        # Mappings (dict and friends)
        if isinstance(a, Mapping) or isinstance(b, Mapping):
            if not (isinstance(a, Mapping) and isinstance(b, Mapping)):
                if debug: print(f"Type mismatch at depth {depth} ({type(a)} vs {type(b)}):\n{a}\nvs\n{b}")
                return False
            if a.keys() != b.keys():
                if debug: print(f"Mapping keys differ at depth {depth}:\n{a.keys()}\nvs\n{b.keys()}")
                return False
            return all(_eq(a[k], b[k], depth + 1) for k in a.keys())

        # Dataclasses (common “object containers”)
        if is_dataclass(a) or is_dataclass(b):
            if not (is_dataclass(a) and is_dataclass(b)):
                if debug: print(f"Type mismatch at depth {depth} ({type(a)} vs {type(b)}):\n{a}\nvs\n{b}")
                return False
            return _eq(asdict(a), asdict(b), depth + 1)

        # Sequence-like custom objects (supports “objects with __iter__/__getitem__” case)
        if _is_sequence_like(a) or _is_sequence_like(b):
            if not (_is_sequence_like(a) and _is_sequence_like(b)):
                if debug: print(f"Type mismatch at depth {depth} ({type(a)} vs {type(b)}):\n{a}\nvs\n{b}")
                return False
            if len(a) != len(b):
                if debug: print(f"Sequence lengths differ at depth {depth} ({len(a)} vs {len(b)}):\n{a} vs {b}")
                return False
            return all(_eq(a[i], b[i], depth + 1) for i in range(len(a)))

        # Iterators/generators: opt-in only (dangerous otherwise)
        if compare_iterators and (isinstance(a, Iterator) or isinstance(b, Iterator)):
            if not (isinstance(a, Iterator) and isinstance(b, Iterator)):
                if debug: print(f"Type mismatch at depth {depth} ({type(a)} vs {type(b)}):\n{a}\nvs\n{b}")
                return False
            # WARNING: consumes iterators
            for aa, bb in zip(a, b):
                if not _eq(aa, bb, depth + 1):
                    if debug: print(f"Iterator values differ at depth {depth}:\n{aa}\nvs\n{bb}")
                    return False
            # ensure same length
            try:
                next(a)
                if debug: print(f"Iterator lengths differ at depth {depth}:\n{a}\nvs\n{b}")
                return False
            except StopIteration:
                pass
            try:
                next(b)
                if debug: print(f"Iterator lengths differ at depth {depth}:\n{a}\nvs\n{b}")
                return False
            except StopIteration:
                pass
            return True

        # Fallback: atomic comparison
        try:
            eq = a == b
            if debug and not eq: print(f"Failed atomic comparison:\n{a}\nvs\n{b}")
            return eq
        except Exception:
            if debug: print(f"Could not perform atomic comparison:\n{a}\nvs\n{b}")
            return False

    return _eq(x, y, 0)



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025                  #
# ######################################### #

import json
import numpy as np

from xtrack.particles.pdg import get_properties_from_pdg_id, is_proton, is_ion

from ...materials import RefMaterial
try:
    from xaux import FsPath  # TODO: once xaux is in Xsuite keep only this
except (ImportError, ModuleNotFoundError):
    from ...xaux import FsPath


_header_start = "! ** XCOLL START  **"
_header_stop  = "! ** XCOLL END  **"


def create_bdsim_config_file(element_dict, particle_ref, physics_list=None, extra_opts=[],
                             extra_input=[], verbose=True, _all_black=False, cwd=None,
                             **kwargs):
    momentum = int(np.ceil((particle_ref.p0c[0] + 1) / 1.e9))  # in GeV
    pdg_id = particle_ref.pdg_id[0]
    if pdg_id is None or pdg_id == 0:
        raise ValueError('particle_ref must have a valid pdg_id')
    particle = pdg_id
    if is_proton(pdg_id):
        particle = '"proton"'
    elif is_ion(pdg_id):
        _, A, Z, _ = get_properties_from_pdg_id(pdg_id)
        q0 = particle_ref.q0
        if not np.isclose(Z, q0):
            if verbose:
                print("Warning: particle_ref is a partially stripped ion.")
        particle = f'"ion {A} {Z} {q0}"'
    else:
        particle = f'"{pdg_id}"'   # Need quoted strings
    gmad = [f"beam, particle={particle}, momentum={momentum}*GeV;"]
    if physics_list is None:
        if is_ion(pdg_id):
            physics_list = "em ftfp_bert hadronic_elastic decay neutron_tracking_cut em_extra ion stopping ion_em_dissociation"
            # physics_list = "em decay hadronic_elastic ftfp_bert stopping neutron_tracking_cut ion ion_em_dissociation"
        else:
            # This is the same as g4FTFP_BERT (verified with option, physicsVerbose=1;)
            # physics_list = "em ftfp_bert hadronic_elastic decay neutron_tracking_cut em_extra ion"
            physics_list = 'g4FTFP_BERT'
    gmad.append(f'option, physicsList="{physics_list}";')
    # gmad.append(f'option, seed={seed};')
    if _all_black:
        gmad.append('option, collimatorsAreInfiniteAbsorbers=1;')
    gmad.append('')
    if extra_opts:
        if hasattr(extra_opts, '__iter__') and not isinstance(extra_opts, str):
            gmad.extend(extra_opts)
        else:
            gmad.append(extra_opts)
        gmad.append('')
    gmad.extend(generate_material_definitions(element_dict, verbose=verbose))
    if extra_input:
        if hasattr(extra_input, '__iter__') and not isinstance(extra_input, str):
            gmad.extend(extra_input)
        else:
            gmad.append(extra_input)
    if cwd is None:
        cwd = FsPath.cwd()
    input_file = FsPath(cwd / 'geant4_input.gmad').resolve()
    with open(input_file, 'w') as fp:
        fp.write('\n'.join([*_generate_xcoll_header(element_dict), *gmad]))
    return input_file, kwargs


def generate_material_definitions(element_dict, verbose=True):
    from ...beam_elements import Geant4CollimatorTip
    code = []
    for name, el in element_dict.items():
        all_mat = [el.material]
        if isinstance(el, Geant4CollimatorTip):
            all_mat.append(el.tip_material)
        for mat in all_mat:
            if mat is None:
                raise ValueError(f"Material not set for element {name}!")
            elif not isinstance(mat, RefMaterial):
                if mat.geant4_name is None:
                    mat._generate_geant4_code()
                if mat._generated_geant4_code is not None \
                and mat._generated_geant4_code not in code:
                    code.append(mat._generated_geant4_code)
            # Verify that material has a Geant4 name (after possible generation)
            if mat.geant4_name is None:
                raise ValueError(f"Material for element {name} has no Geant4 name!")
        if verbose:
            mats = [m.geant4_name for m in all_mat]
            if len(mats) == 1:
                print(f"Element {name} (id {el.geant4_id}) uses material {mats[0]}")
            else:
                print(f"Element {name} (id {el.geant4_id}) uses materials {mats}")
    code.append('')
    return code


def get_collimators_from_input_file(input_file):
    with open(input_file, 'r') as fp:
        data = fp.read()
    if _header_start not in data or _header_stop not in data:
        raise ValueError("No Xcoll header found in input file. Regenerate input file!")
    commented_dict = data.split(_header_start)[1].split(_header_stop)[0].split('\n')[1:-1]
    cleaned_dict = "".join([val[3:] for val in commented_dict])
    return json.loads(cleaned_dict)


def _generate_xcoll_header(element_dict):
    from ...beam_elements import Geant4CollimatorTip
    header = ["!  DO NOT CHANGE THIS HEADER", _header_start, "!  {"]
    for kk, el in element_dict.items():
        vv = {
            "geant4_id": np.array(el.geant4_id).tolist(),
            "length": np.array(el.length).tolist(),
            "material": np.array(el.material.geant4_name).tolist(),
            "angle": np.array(el.angle).tolist(),
            "jaw": [np.array(el.jaw_L).tolist(), np.array(el.jaw_R).tolist()],
            "tilt": np.array(el.tilt).tolist(),
        }
        if isinstance(el, Geant4CollimatorTip):
            vv["tip_material"] = np.array(el.tip_material.geant4_name).tolist()
            vv["tip_thickness"]   = np.array(el.tip_thickness).tolist()
        header.append(f'!  "{kk}": ' + json.dumps(vv) + ',')
    header[-1] = header[-1][:-1]  # remove last comma
    header.append("!  }")
    header.append(_header_stop)
    header.append('')
    return header


# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from math import sqrt

import xtrack as xt
from xtrack.particles.pdg import get_name_from_pdg_id, get_properties_from_pdg_id, is_ion
try:
    from xaux import FsPath  # TODO: once xaux is in Xsuite keep only this
except (ImportError, ModuleNotFoundError):
    from ...xaux import FsPath

from .environment import format_fluka_float
from .prototype import FlukaPrototype, FlukaAssembly
from ...general import _pkg_root


def _is_ion(pdg_id):
    if isinstance(pdg_id, xt.Particles):
        pdg_id = pdg_id.pdg_id[0]
    return is_ion(pdg_id)


def get_include_files(particle_ref, include_files=[], *, verbose=True, assemblies=[],
                      bb_int=False, touches=False, **kwargs):

    import xcoll as xc
    phys = xc.fluka.engine._physics_settings
    this_include_files = include_files.copy()

    # Required default include files
    if 'include_settings_beam.inp' not in [file.name for file in this_include_files]:
        this_include_files.append(_beam_include_file(particle_ref, bb_int=bb_int))
    if 'include_settings_physics.inp' not in [file.name for file in this_include_files]:
        physics_file = _physics_include_file(verbose=verbose, lower_momentum_cut=phys.hadron_lower_momentum_cut,
                                             photon_lower_momentum_cut=phys.photon_lower_momentum_cut,
                                             electron_lower_momentum_cut=phys.electron_lower_momentum_cut,
                                             include_showers=phys.include_showers)
        this_include_files.append(physics_file)
    if 'include_custom_scoring.inp' not in [file.name for file in this_include_files]:
        scoring_file = _scoring_include_file(verbose=verbose, return_list=phys, get_touches=touches,
                                             use_crystals=any([assm.is_crystal for assm in assemblies]))
        this_include_files.append(scoring_file)
    if 'include_custom_assignmat.inp' not in [file.name for file in this_include_files]:
        material_file = _assignmat_include_file(assemblies=assemblies)
        this_include_files.append(material_file)

    # Add any additional include files
    if 'include_custom_biasing.inp' not in [file.name for file in this_include_files]:
        this_include_files.append(_biasing_include_file())
    if 'include_define.inp' not in [file.name for file in this_include_files]:
        this_include_files.append(_define_include_file())
    this_include_files = [FsPath(ff).resolve() for ff in this_include_files]
    for ff in this_include_files:
        if not ff.exists():
            raise FileNotFoundError(f"Include file not found: {ff}.")
        elif ff.parent != FsPath.cwd():
            ff.copy_to(FsPath.cwd(), method='mount')
    return this_include_files, kwargs


def _assignmat_include_file(assemblies=[]):
    from xcoll.materials.database import db as mdb
    template =  ''.join([mat._generated_fluka_code for mat in mdb.fluka.values()
                         if mat._generated_fluka_code is not None])
    if any([assm.is_crystal for assm in assemblies]):
        template += f"""\
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
* Crystal Card
* what (1) = REGNUM ( mandatory )
* what (2) = Crystal bending angle [ mrad ]
* what (3) = A parameter = Crystal length [ cm ]
* what (4) = B parameter = Torsion coefficient [ mrad / cm ]
* what (5) = C parameter = Quasi - Mosaic factor [ mrad / cm ]
* what (6) = D parameter = Crystal temperature [K]
* sdum = crystal type
* what (7, 8, 9) = U/V/ W of nominal crystal normal to the curvature plane
* what (10, 11, 12) = U /V/W of nominal crystal channel direction
* what (13, 14, 15) = X /Y/Z of crystal nominal position
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
* CRYSTAL what (1) what (2) what (3) what (4) what (5) what (6) sdum
* CRYSTAL what (7) what (8) what (9) what (10) what (11) what (12) &
* CRYSTAL what (13) what (14) what (15) &&
*
"""
    for crystal in assemblies:
        if not crystal.is_crystal:
            continue
        name = crystal.name
        if isinstance(crystal, FlukaAssembly) and crystal.is_generic():
            name = crystal.prototypes[1].name   # region is tied to crystal body
        l = crystal.length * 100
        bang = round(l/(crystal.bending_radius*100) *1000, 6) # mrad
        if crystal.fluka_position is not None:
            pos = crystal.fluka_position
        else:
            if isinstance(crystal, FlukaAssembly):
                raise ValueError(f"Crystal {name} has no position in the prototypes file.")
            if len(crystal.dependant_assemblies) == 0:
                raise ValueError(f"Crystal {name} has no position in the prototypes file.")
            elif len(crystal.dependant_assemblies) > 1:
                raise ValueError(f"Crystal {name} has multiple positions in the prototypes file. "
                                + "(too many dependant asemblies).")
            pos = crystal.dependant_assemblies[0].fluka_position
        pos1   = format_fluka_float(pos[3]-50.0+1e-4) # XXX Does it always work with the 50cm shift?
        pos2   = format_fluka_float(pos[4])
        pos3   = format_fluka_float(pos[5])
        template += f"""\
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
CRYSTAL     {name:>8}{format_fluka_float(bang)}{format_fluka_float(l)}       0.0       0.0     300.0 110
CRYSTAL          0.0      -1.0       0.0       0.0       0.0       1.0 &
CRYSTAL   {pos1}{pos2}{pos3}                              &&
"""
    filename = FsPath("include_custom_assignmat.inp").resolve()

    with filename.open('w') as fp:
        fp.write(template)
    return filename


def _beam_include_file(particle_ref, bb_int=False):
    import xcoll as xc
    filename = FsPath("include_settings_beam.inp").resolve()
    pdg_id = particle_ref.pdg_id[0]
    momentum_cut = particle_ref.p0c[0] / 1e9 * 1.05

    hi_prope = "*"
    if pdg_id in xc.fluka.particle_names:
        name = xc.fluka.particle_names[pdg_id]
    elif _is_ion(pdg_id):
        _, A, Z, _ = get_properties_from_pdg_id(pdg_id)
        momentum_cut *= 3.2 / A # Upper limit (scaling is slightly arbitrary)
        name = "HEAVYION"
        hi_prope = f"HI-PROPE  {Z:8}.0{A:8}.0"
    else:
        raise ValueError(f"Reference particle {get_name_from_pdg_id(pdg_id)} not "
                        + "supported by FLUKA.")
    beam = f"BEAM      {format_fluka_float(momentum_cut)}{50*' '}{name}"
    template = f"""\
******************************************************************************
*                          BEAM SETTINGS                                     *
******************************************************************************
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
* maximum momentum per nucleon (3000 for 3.5Z TeV, 6000 for 6.37Z TeV)
{beam}
{hi_prope}
*
BEAMPOS
*
"""
    if bb_int:
        template += f"""\
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
* W(1): interaction type, W(2): Index IR, W(3): Length of IR[cm], W(4): SigmaZ
* Interaction type: 1.0 (Inelastic), 10.0 (Elastic), 100.0 (EMD)
* SigmaZ: sigma_z (cm) for the Gaussian sampling of the collision position around the center of the insertion
SOURCE    {format_fluka_float(bb_int['int_type'])}{format_fluka_float(1)}{format_fluka_float(20)}{format_fluka_float(bb_int['sigma_z'])}
SOURCE           89.       90.       91.        0.                    &
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
* W(1): Theta2-> Polar angle (rad) between b2 and -z direction.
* W(2): Azimuthal angle (deg) defining the crossing plane.
* W(3): Sigma_x for Gaussian sampling in b2, dir (x', y', 1) wrt to px
* W(4): Sigma_y for Gaussian sampling in b2, dir (x', y', 1) wrt to py
* W(5): Z b2
* W(6): A b2
USRICALL  {format_fluka_float(bb_int['theta2'])}{format_fluka_float(bb_int['xs'])}{format_fluka_float(bb_int['sigma_p_x2'])}{format_fluka_float(bb_int['sigma_p_y2'])}{format_fluka_float(bb_int['Z'])}{format_fluka_float(bb_int['A'])}BBEAMCOL
USRICALL  {format_fluka_float(bb_int['betx'])}{format_fluka_float(bb_int['alfx'])}{format_fluka_float(bb_int['dx'])}{format_fluka_float(bb_int['dpx'])}{format_fluka_float(bb_int['rms_emx'])}          BBCOL_H
USRICALL  {format_fluka_float(bb_int['bety'])}{format_fluka_float(bb_int['alfy'])}{format_fluka_float(bb_int['dy'])}{format_fluka_float(bb_int['dpy'])}{format_fluka_float(bb_int['rms_emy'])}          BBCOL_V
* * ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
* *USRICALL      #OFFSET                                                        BBCOL_O
* W(1,2,3,4,5,6): X, X', Y,  Y', dT, pc
* USRICALL         0.0       0.0       0.0       0.0       0.0       0.0BBCOL_O
USRICALL  {format_fluka_float(bb_int['offset'][0])}{format_fluka_float(bb_int['offset'][1])}{format_fluka_float(bb_int['offset'][2])}{format_fluka_float(bb_int['offset'][3])}       0.0       0.0BBCOL_O
* *USRICALL      #SIGDPP                                                        BBCOL_L
* USRICALL         0.0       0.0       0.0       0.0       0.0       0.0BBCOL_L
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
USRICALL  {format_fluka_float(bb_int['sigma_dpp'])}                                                  BBCOL_L
*
"""
    else:
        template += f"""\
* Only asking for loss map and touches map as in Xsuite
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
SOURCE                                         87.       88.        1.
SOURCE           89.       90.       91.        0.       -1.       10.&
SOURCE           0.0       0.0      97.0       1.0      96.0       1.0&&
"""


    with filename.open('w') as fp:
        fp.write(template)
    return filename


def _physics_include_file(*, verbose, lower_momentum_cut, photon_lower_momentum_cut,
                          electron_lower_momentum_cut, include_showers):
    filename = FsPath("include_settings_physics.inp").resolve()
    emf = "*EMF" if include_showers else "EMF"
    deltaray = "DELTARAY" if not include_showers else "*DELTARAY"
    emfcut = "EMFCUT" if include_showers else "*EMFCUT"
    photon_lower_momentum_cut = format_fluka_float(photon_lower_momentum_cut/1.e9)
    # TODO: FLUKA electron mass
    electron_lower_energy_cut = sqrt(electron_lower_momentum_cut**2 + 511e3**2)
    electron_lower_energy_cut = format_fluka_float(electron_lower_energy_cut/1.e9)
    lower_momentum_cut /= 1.e9
    if verbose:
        print(f"Physics include file created with:\n"
             + f"  - Hadron and muon lower momentum cut: {lower_momentum_cut} GeV")
        if include_showers:
            print(f"  - EM showers: ON\n"
                + f"  - Photon lower momentum cut: {photon_lower_momentum_cut} GeV\n"
                + f"  - Electron lower momentum cut: {electron_lower_energy_cut} GeV")
        else:
            print(f"  - EM showers: OFF")

    template = f"""\
******************************************************************************
*                          PHYSICS SETTINGS                                  *
******************************************************************************
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
DEFAULTS                                                              PRECISIO
*
* Thresholds for secondary particle (electron, positron, photon) production
* applied to all materials; electron, positron: 1.0 MeV, photons: 0.1 MeV
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
{emfcut}    {electron_lower_energy_cut}{photon_lower_momentum_cut}       1.0       1.0  @LASTMAT       1.0PROD-CUT
{emfcut}    {electron_lower_energy_cut}{photon_lower_momentum_cut}       0.0       1.0  @LASTREG       1.0
*
* Kill EM showers
{emf}                                                                   EMF-OFF
{deltaray}          -1                      BLCKHOLE  @LASTMAT
*
* All particle transport thresholds up to 1 TeV
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
PART-THR  {format_fluka_float(lower_momentum_cut)}            @LASTPAR                 0.0
PART-THR  {format_fluka_float(2*lower_momentum_cut)}  DEUTERON                           0.0
PART-THR  {format_fluka_float(3*lower_momentum_cut)}    TRITON                           0.0
PART-THR  {format_fluka_float(3*lower_momentum_cut)}  3-HELIUM                           0.0
PART-THR  {format_fluka_float(4*lower_momentum_cut)}  4-HELIUM                           0.0
*
* Activate single scattering
MULSOPT                                        1.0       1.0       1.0GLOBAL
*
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
* un-comment in case you use FLUKA pro
PHYSICS           1.                                                  COALESCE
PHYSICS           3.                                                  EVAPORAT
PHYSICS        1.D+5     1.D+5     1.D+5     1.D+5     1.D+5     1.D+5PEATHRES
PHYSICS           2.                                                  EM-DISSO
* beam-beam collisions
PHYSICS           -1                                                  LIMITS
* No low-energy neutron transport
LOW-PWXS          -1
"""
    with filename.open('w') as fp:
        fp.write(template)
    return filename


def _scoring_include_file(*, verbose, return_list, get_touches=False, use_crystals=False):
    filename = FsPath("include_custom_scoring.inp").resolve()
    all_charged = 'USRBDX' if return_list.return_all_charged else '*USRBDX'
    all_particles = 'USRBDX' if return_list.return_all else '*USRBDX'
    neutral = return_list.return_neutral
    allow = not return_list.return_all_charged and not return_list.return_all
    photons = 'USRBDX' if return_list.return_photons and allow else '*USRBDX'
    electrons = 'USRBDX' if return_list.return_electrons and allow else '*USRBDX'
    muons = 'USRBDX' if return_list.return_muons and allow else '*USRBDX'
    tauons = 'USRBDX' if return_list.return_tauons and allow else '*USRBDX'
    neutrinos = 'USRBDX' if return_list.return_neutrinos and allow else '*USRBDX'
    protons = 'USRBDX' if return_list.return_protons and allow else '*USRBDX'
    neutrons = 'USRBDX' if return_list.return_neutrons and allow else '*USRBDX'
    baryons = 'USRBDX' if return_list.return_other_baryons and allow else '*USRBDX'
    neutral_baryons = 'USRBDX' if return_list.return_other_baryons and neutral and allow else '*USRBDX'
    pions = 'USRBDX' if return_list.return_pions and allow else '*USRBDX'
    neutral_pions = 'USRBDX' if return_list.return_pions and neutral and allow else '*USRBDX'
    kaons = 'USRBDX' if return_list.return_kaons and allow else '*USRBDX'
    neutral_kaons = 'USRBDX' if return_list.return_kaons and neutral and allow else '*USRBDX'
    mesons = 'USRBDX' if return_list.return_other_mesons and allow else '*USRBDX'
    neutral_mesons = 'USRBDX' if return_list.return_other_mesons and neutral and allow else '*USRBDX'
    ions = 'USRBDX' if return_list.return_ions and allow else '*USRBDX'
    crystal = 'USRICALL' if use_crystals else '*USRICALL'
    touches = 'USERDUMP' if get_touches else '*USERDUMP'
    if verbose:
        print(f"Scoring include file created with:")
        if all_particles[0] != '*':
            print(f"  - Return all particles: True")
        elif all_charged[0] != '*':
            print(f"  - Return all charged particles: True")
        else:
            print(f"  - Return photons: {photons[0] != '*'}\n"
                + f"  - Return electrons: {electrons[0] != '*'}\n"
                + f"  - Return muons: {muons[0] != '*'}\n"
                + f"  - Return tauons: {tauons[0] != '*'}\n"
                + f"  - Return neutrinos: {neutrinos[0] != '*'}\n"
                + f"  - Return protons: {protons[0] != '*'}\n"
                + f"  - Return neutrons: {neutrons[0] != '*'}\n"
                + f"  - Return other baryons: {baryons[0] != '*'}\n"
                + f"  - Return neutral baryons: {neutral_baryons[0] != '*'}\n"
                + f"  - Return pions: {pions[0] != '*'}\n"
                + f"  - Return neutral pions: {neutral_pions[0] != '*'}\n"
                + f"  - Return kaons: {kaons[0] != '*'}\n"
                + f"  - Return neutral kaons: {neutral_kaons[0] != '*'}\n"
                + f"  - Return other mesons: {mesons[0] != '*'}\n"
                + f"  - Return neutral mesons: {neutral_mesons[0] != '*'}\n"
                + f"  - Return ions: {ions[0] != '*'}\n")
        print(f"  - Use crystals: {crystal[0] != '*'}")
        print(f"  - Get touches map: {touches[0] != '*'}")

    template = f"""\
* New way to give back particles to Icosim via fluscw.f routine (no more usrmed)
*     through a fake USRBDX estimator
USERWEIG                             3.0
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7....+....8
{all_charged}          99.0  ALL-CHAR     -42.0   VAROUND  TRANSF_D          BACK2ICO
{all_particles}          99.0  ALL-PART     -42.0   VAROUND  TRANSF_D          BACK2ICO
{photons}          99.0    PHOTON     -42.0   VAROUND  TRANSF_D          BACK2ICO
{photons}          99.0  OPTIPHOT     -42.0   VAROUND  TRANSF_D          BACK2ICO
{photons}          99.0       RAY     -42.0   VAROUND  TRANSF_D          BACK2ICO
{electrons}          99.0  ELECTRON     -42.0   VAROUND  TRANSF_D          BACK2ICO
{electrons}          99.0  POSITRON     -42.0   VAROUND  TRANSF_D          BACK2ICO
{muons}          99.0     MUON+     -42.0   VAROUND  TRANSF_D          BACK2ICO
{muons}          99.0     MUON-     -42.0   VAROUND  TRANSF_D          BACK2ICO
{tauons}          99.0      TAU+     -42.0   VAROUND  TRANSF_D          BACK2ICO
{tauons}          99.0      TAU-     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutrinos}          99.0   NEUTRIE     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutrinos}          99.0  ANEUTRIE     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutrinos}          99.0   NEUTRIM     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutrinos}          99.0  ANEUTRIM     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutrinos}          99.0   NEUTRIT     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutrinos}          99.0  ANEUTRIT     -42.0   VAROUND  TRANSF_D          BACK2ICO
{pions}          99.0     PION+     -42.0   VAROUND  TRANSF_D          BACK2ICO
{pions}          99.0     PION-     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_pions}          99.0    PIZERO     -42.0   VAROUND  TRANSF_D          BACK2ICO
{kaons}          99.0     KAON+     -42.0   VAROUND  TRANSF_D          BACK2ICO
{kaons}          99.0     KAON-     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_kaons}          99.0  KAONZERO     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_kaons}          99.0  AKAONZER     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_kaons}          99.0  KAONLONG     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_kaons}          99.0  KAONSHRT     -42.0   VAROUND  TRANSF_D          BACK2ICO
{protons}          99.0    PROTON     -42.0   VAROUND  TRANSF_D          BACK2ICO
{protons}          99.0   APROTON     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutrons}          99.0   NEUTRON     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutrons}          99.0  ANEUTRON     -42.0   VAROUND  TRANSF_D          BACK2ICO
{ions}          99.0  DEUTERON     -42.0   VAROUND  TRANSF_D          BACK2ICO
{ions}          99.0    TRITON     -42.0   VAROUND  TRANSF_D          BACK2ICO
{ions}          99.0  3-HELIUM     -42.0   VAROUND  TRANSF_D          BACK2ICO
{ions}          99.0  4-HELIUM     -42.0   VAROUND  TRANSF_D          BACK2ICO
{ions}          99.0  HEAVYION     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_baryons}          99.0    LAMBDA     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_baryons}          99.0   ALAMBDA     -42.0   VAROUND  TRANSF_D          BACK2ICO
{baryons}          99.0  LAMBDAC+     -42.0   VAROUND  TRANSF_D          BACK2ICO
{baryons}          99.0  ALAMBDC-     -42.0   VAROUND  TRANSF_D          BACK2ICO
{baryons}          99.0    SIGMA-     -42.0   VAROUND  TRANSF_D          BACK2ICO
{baryons}          99.0    SIGMA+     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_baryons}          99.0  SIGMAZER     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_baryons}          99.0  ASIGMAZE     -42.0   VAROUND  TRANSF_D          BACK2ICO
{baryons}          99.0   ASIGMA-     -42.0   VAROUND  TRANSF_D          BACK2ICO
{baryons}          99.0   ASIGMA+     -42.0   VAROUND  TRANSF_D          BACK2ICO
{baryons}          99.0      XSI-     -42.0   VAROUND  TRANSF_D          BACK2ICO
{baryons}          99.0     AXSI+     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_baryons}          99.0   XSIZERO     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_baryons}          99.0  AXSIZERO     -42.0   VAROUND  TRANSF_D          BACK2ICO
{baryons}          99.0     XSIC+     -42.0   VAROUND  TRANSF_D          BACK2ICO
{baryons}          99.0    AXSIC-     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_baryons}          99.0     XSIC0     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_baryons}          99.0    AXSIC0     -42.0   VAROUND  TRANSF_D          BACK2ICO
{baryons}          99.0    XSIPC+     -42.0   VAROUND  TRANSF_D          BACK2ICO
{baryons}          99.0   AXSIPC-     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_baryons}          99.0    XSIPC0     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_baryons}          99.0   AXSIPC0     -42.0   VAROUND  TRANSF_D          BACK2ICO
{baryons}          99.0    OMEGA-     -42.0   VAROUND  TRANSF_D          BACK2ICO
{baryons}          99.0   AOMEGA+     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_baryons}          99.0   OMEGAC0     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_baryons}          99.0  AOMEGAC0     -42.0   VAROUND  TRANSF_D          BACK2ICO
{mesons}          99.0        D+     -42.0   VAROUND  TRANSF_D          BACK2ICO
{mesons}          99.0        D-     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_mesons}          99.0        D0     -42.0   VAROUND  TRANSF_D          BACK2ICO
{neutral_mesons}          99.0     D0BAR     -42.0   VAROUND  TRANSF_D          BACK2ICO
{mesons}          99.0       DS+     -42.0   VAROUND  TRANSF_D          BACK2ICO
{mesons}          99.0       DS-     -42.0   VAROUND  TRANSF_D          BACK2ICO
*
* Get back touches
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7....+....8
{touches}       100.0
*
* crystal scoring
{crystal}        50.0                                                  CRYSTAL
"""
    with filename.open('w') as fp:
        fp.write(template)
    return filename


def _biasing_include_file():
    filename = FsPath(f"include_custom_biasing.inp").resolve()
    template = f"""\
******************************************************************************
*                             BIASING SETTINGS                               *
******************************************************************************
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
*
* Place in this file all the custom biasing you require
*
* --------------------------------
* Standard Leading particle biasing applies everywhere
*________+_________+_________+_________+_________+_________+_________+
*EMF-BIAS      1022.0     10.00     10.00       2.0  @LASTREG          LPBEMF
*
"""
    with filename.open('w') as fp:
        fp.write(template)
    return filename


def _define_include_file():
    filename = FsPath(f"include_define.inp").resolve()
    template = f"""\
******************************************************************************
*                                  DEFINES                                   *
******************************************************************************
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
*
* Place in this file all the custom defines you require
*
"""
    with filename.open('w') as fp:
        fp.write(template)
    return filename



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from xtrack.particles import pdg

from .material import Material
from .database import db, _manually_add_material_to_db


# Atomic mass from https://iupac.qmul.ac.uk/AtWt/ rounded to 5 digits
# Densities (in g/cm3) from:
#     https://en.wikipedia.org/wiki/List_of_chemical_elements
#     https://en.wikipedia.org/wiki/Densities_of_the_elements_(data_page)
Hydrogen      = Material(Z=1,     A=1.0080,   state='gas',     density=89.88e-6)
Helium        = Material(Z=2,     A=4.0026,   state='gas',     density=178.5e-6)
Lithium       = Material(Z=3,     A=6.94,     state='solid',   density=0.534)
Beryllium     = Material(Z=4,     A=9.0122,   state='solid',   density=1.848)
Boron         = Material(Z=5,     A=10.81,    state='solid',   density=2.34)
Carbon        = Material(Z=6,     A=12.011,   state='solid',   density=2.265)   # Amorphous
Nitrogen      = Material(Z=7,     A=14.007,   state='gas',     density=0.0012506)
Oxygen        = Material(Z=8,     A=15.999,   state='gas',     density=0.001429)
Fluorine      = Material(Z=9,     A=18.998,   state='gas',     density=0.001696)
Neon          = Material(Z=10,    A=20.180,   state='gas',     density=0.0009002)
Sodium        = Material(Z=11,    A=22.990,   state='solid',   density=0.968)
Magnesium     = Material(Z=12,    A=24.305,   state='solid',   density=1.738)
Aluminium     = Material(Z=13,    A=26.982,   state='solid',   density=2.70)
Silicon       = Material(Z=14,    A=28.085,   state='solid',   density=2.3296)
Phosphorus    = Material(Z=15,    A=30.974,   state='solid',   density=2.2)     # Red phosphorus
Sulfur        = Material(Z=16,    A=32.06,    state='solid',   density=2.067)
Chlorine      = Material(Z=17,    A=35.45,    state='gas',     density=0.003214)
Argon         = Material(Z=18,    A=39.948,   state='gas',     density=0.001784)
Potassium     = Material(Z=19,    A=39.098,   state='solid',   density=0.862)
Calcium       = Material(Z=20,    A=40.078,   state='solid',   density=1.54)
Scandium      = Material(Z=21,    A=44.956,   state='solid',   density=2.99)
Titanium      = Material(Z=22,    A=47.867,   state='solid',   density=4.506)
Vanadium      = Material(Z=23,    A=50.942,   state='solid',   density=6.11)
Chromium      = Material(Z=24,    A=51.996,   state='solid',   density=7.19)
Manganese     = Material(Z=25,    A=54.938,   state='solid',   density=7.26)
Iron          = Material(Z=26,    A=55.845,   state='solid',   density=7.874)
Cobalt        = Material(Z=27,    A=58.933,   state='solid',   density=8.90)
Nickel        = Material(Z=28,    A=58.693,   state='solid',   density=8.908)
Copper        = Material(Z=29,    A=63.546,   state='solid',   density=8.96)
Zinc          = Material(Z=30,    A=65.38,    state='solid',   density=7.134)
Gallium       = Material(Z=31,    A=69.723,   state='solid',   density=5.91)
Germanium     = Material(Z=32,    A=72.63,    state='solid',   density=5.323)
Arsenic       = Material(Z=33,    A=74.922,   state='solid',   density=5.727)
Selenium      = Material(Z=34,    A=78.971,   state='solid',   density=4.809)
Bromine       = Material(Z=35,    A=79.904,   state='liquid',  density=3.1028)
Krypton       = Material(Z=36,    A=83.798,   state='gas',     density=0.003749)
Rubidium      = Material(Z=37,    A=85.468,   state='solid',   density=1.532)
Strontium     = Material(Z=38,    A=87.62,    state='solid',   density=2.64)
Yttrium       = Material(Z=39,    A=88.906,   state='solid',   density=4.469)
Zirconium     = Material(Z=40,    A=91.224,   state='solid',   density=6.506)
Niobium       = Material(Z=41,    A=92.906,   state='solid',   density=8.57)
Molybdenum    = Material(Z=42,    A=95.95,    state='solid',   density=10.223)
Technetium    = Material(Z=43,    A=97,       state='solid',   density=11.5)    # Only unstable isotopes
Ruthenium     = Material(Z=44,    A=101.07,   state='solid',   density=12.37)
Rhodium       = Material(Z=45,    A=102.91,   state='solid',   density=12.41)
Palladium     = Material(Z=46,    A=106.42,   state='solid',   density=12.02)
Silver        = Material(Z=47,    A=107.87,   state='solid',   density=10.49)
Cadmium       = Material(Z=48,    A=112.41,   state='solid',   density=8.65)
Indium        = Material(Z=49,    A=114.82,   state='solid',   density=7.31)
Tin           = Material(Z=50,    A=118.71,   state='solid',   density=7.365)
Antimony      = Material(Z=51,    A=121.76,   state='solid',   density=6.697)
Tellurium     = Material(Z=52,    A=127.60,   state='solid',   density=6.24)
Iodine        = Material(Z=53,    A=126.90,   state='solid',   density=4.933)
Xenon         = Material(Z=54,    A=131.29,   state='gas',     density=0.005894)
Cesium        = Material(Z=55,    A=132.91,   state='solid',   density=1.93)
Barium        = Material(Z=56,    A=137.33,   state='solid',   density=3.62)
Lanthanum     = Material(Z=57,    A=138.91,   state='solid',   density=6.15)
Cerium        = Material(Z=58,    A=140.12,   state='solid',   density=6.77)
Praseodymium  = Material(Z=59,    A=140.91,   state='solid',   density=6.77)
Neodymium     = Material(Z=60,    A=144.24,   state='solid',   density=7.01)
Promethium    = Material(Z=61,    A=145,      state='solid',   density=7.26)    # Only unstable isotopes
Samarium      = Material(Z=62,    A=150.36,   state='solid',   density=7.52)
Europium      = Material(Z=63,    A=151.96,   state='solid',   density=5.24)
Gadolinium    = Material(Z=64,    A=157.25,   state='solid',   density=7.90)
Terbium       = Material(Z=65,    A=158.93,   state='solid',   density=8.23)
Dysprosium    = Material(Z=66,    A=162.50,   state='solid',   density=8.55)
Holmium       = Material(Z=67,    A=164.93,   state='solid',   density=8.80)
Erbium        = Material(Z=68,    A=167.26,   state='solid',   density=9.07)
Thulium       = Material(Z=69,    A=168.93,   state='solid',   density=9.32)
Ytterbium     = Material(Z=70,    A=173.05,   state='solid',   density=6.97)
Lutetium      = Material(Z=71,    A=174.97,   state='solid',   density=9.84)
Hafnium       = Material(Z=72,    A=178.49,   state='solid',   density=13.31)
Tantalum      = Material(Z=73,    A=180.95,   state='solid',   density=16.65)
Tungsten      = Material(Z=74,    A=183.84,   state='solid',   density=19.3)
Rhenium       = Material(Z=75,    A=186.21,   state='solid',   density=21.02)
Osmium        = Material(Z=76,    A=190.23,   state='solid',   density=22.59)
Iridium       = Material(Z=77,    A=192.22,   state='solid',   density=22.56)
Platinum      = Material(Z=78,    A=195.08,   state='solid',   density=21.45)
Gold          = Material(Z=79,    A=196.97,   state='solid',   density=19.32)
Mercury       = Material(Z=80,    A=200.59,   state='liquid',  density=13.534)
Thallium      = Material(Z=81,    A=204.38,   state='solid',   density=11.85)
Lead          = Material(Z=82,    A=207.19,   state='solid',   density=11.348)
Bismuth       = Material(Z=83,    A=208.98,   state='solid',   density=9.78)
Polonium      = Material(Z=84,    A=209,      state='solid',   density=9.32)    # Only unstable isotopes
Astatine      = Material(Z=85,    A=210,      state='solid',   density=7.0)     # Only unstable isotopes
Radon         = Material(Z=86,    A=222,      state='gas',     density=0.00973) # Only unstable isotopes
Francium      = Material(Z=87,    A=223,      state='solid',   density=2.48)    # Only unstable isotopes
Radium        = Material(Z=88,    A=226,      state='solid',   density=5.50)    # Only unstable isotopes
Actinium      = Material(Z=89,    A=227,      state='solid',   density=10.07)   # Only unstable isotopes
Thorium       = Material(Z=90,    A=232.04,   state='solid',   density=11.72)
Protactinium  = Material(Z=91,    A=231.04,   state='solid',   density=15.37)
Uranium       = Material(Z=92,    A=238.03,   state='solid',   density=18.95)
Neptunium     = Material(Z=93,    A=237,      state='solid',   density=20.45)   # Only unstable isotopes
Plutonium     = Material(Z=94,    A=244,      state='solid',   density=19.84)   # Only unstable isotopes
Americium     = Material(Z=95,    A=243,      state='solid',   density=13.69)   # Only unstable isotopes
Curium        = Material(Z=96,    A=247,      state='solid',   density=13.51)   # Only unstable isotopes
Berkelium     = Material(Z=97,    A=247,      state='solid',   density=14.78)   # Only unstable isotopes
Californium   = Material(Z=98,    A=251,      state='solid',   density=15.1)    # Only unstable isotopes
Einsteinium   = Material(Z=99,    A=252,      state='solid',   density=8.84)    # Only unstable isotopes
Fermium       = Material(Z=100,   A=257,      state='solid',   density=9.7)     # Only unstable isotopes
Mendelevium   = Material(Z=101,   A=258,      state='solid',   density=10.3)    # Only unstable isotopes
Nobelium      = Material(Z=102,   A=259,      state='solid',   density=9.9)     # Only unstable isotopes
Lawrencium    = Material(Z=103,   A=266,      state='solid',   density=15.6)    # Only unstable isotopes
Rutherfordium = Material(Z=104,   A=267,      state='solid',   density=23.2)    # Only unstable isotopes
Dubnium       = Material(Z=105,   A=268,      state='solid',   density=29.3)    # Only unstable isotopes
Seaborgium    = Material(Z=106,   A=269,      state='solid',   density=35.0)    # Only unstable isotopes
Bohrium       = Material(Z=107,   A=270,      state='solid',   density=37.1)    # Only unstable isotopes
Hassium       = Material(Z=108,   A=277,      state='solid',   density=41.0)    # Only unstable isotopes
Meitnerium    = Material(Z=109,   A=278,      state='solid',   density=37.4)    # Only unstable isotopes
Darmstadtium  = Material(Z=110,   A=281,      state='solid',   density=34.8)    # Only unstable isotopes
Roentgenium   = Material(Z=111,   A=282,      state='solid',   density=28.7)    # Only unstable isotopes
Copernicium   = Material(Z=112,   A=285,      state='solid',   density=14.0)    # Only unstable isotopes
Nihonium      = Material(Z=113,   A=286,      state='solid',   density=16.0)    # Only unstable isotopes
Flerovium     = Material(Z=114,   A=289,      state='solid',   density=14.0)    # Only unstable isotopes
Moscovium     = Material(Z=115,   A=290,      state='solid',   density=13.5)    # Only unstable isotopes
Livermorium   = Material(Z=116,   A=293,      state='solid',   density=12.9)    # Only unstable isotopes
Tennessine    = Material(Z=117,   A=294,      state='solid',   density=7.2)     # Only unstable isotopes
Oganesson     = Material(Z=118,   A=294,      state='solid',   density=5.0)     # Only unstable isotopes


# Extra parameters for Everest
# ============================

Beryllium.adapt(inplace=True,  nuclear_radius=0.22, nuclear_elastic_slope=74.7,
                               cross_section=[0.271, 0.192, 0, 0, 0, 0.0035e-2], hcut=0.02)
Aluminium.adapt(inplace=True,  nuclear_radius=0.302, nuclear_elastic_slope=120.3,
                               cross_section=[0.643, 0.418, 0, 0, 0, 0.0340e-2], hcut=0.02)
Silicon.adapt(inplace=True,    nuclear_radius=0.441, nuclear_elastic_slope=120.14,
                               cross_section=[0.664, 0.430, 0, 0, 0, 0.0390e-2], hcut=0.02,
                               crystal_plane_distance=0.96e-7, crystal_potential=21.34,
                               nuclear_collision_length=0.3016, eta=0.9)
Copper.adapt(inplace=True,     nuclear_radius=0.366, nuclear_elastic_slope=217.8,
                               cross_section=[1.253, 0.769, 0, 0, 0, 0.1530e-2], hcut=0.01)
Germanium.adapt(inplace=True,  nuclear_radius=0.605, nuclear_elastic_slope=226.35,
                               cross_section=[1.388, 0.844, 0, 0, 0, 0.1860e-2], hcut=0.02,
                               crystal_plane_distance=1.0e-7, crystal_potential=40.0,
                               nuclear_collision_length=0.1632, eta=0.9)
Molybdenum.adapt(inplace=True, nuclear_radius=0.481, nuclear_elastic_slope=273.9,
                               cross_section=[1.713, 1.023, 0, 0, 0, 0.2650e-2], hcut=0.02)
Tungsten.adapt(inplace=True,   nuclear_radius=0.520, nuclear_elastic_slope=440.3,
                               cross_section=[2.765, 1.591, 0, 0, 0, 0.7680e-2], hcut=0.01,
                               crystal_plane_distance=0.56e-7, crystal_potential=21.0,
                               nuclear_collision_length=1.e-12, eta=0.9)
Lead.adapt(inplace=True,       nuclear_radius=0.542, nuclear_elastic_slope=455.3,
                               cross_section=[3.016, 1.724, 0, 0, 0, 0.9070e-2], hcut=0.01)


# Metadata for database
# =====================

for name, obj in list(globals().items()):  # Have to wrap in list to take a snapshot (avoid updating in-place)
    if isinstance(obj, Material) and obj.name is None:
        # Check consistency between name and Z
        if name == 'Aluminium':
            assert obj.Z == 13
        else:
            assert pdg.get_element_full_name_from_Z(obj.Z).lower() == name.lower(), \
            f'Inconsistency between material name {name} and Z={obj.Z} ({pdg.get_element_full_name_from_Z(obj.Z)})'

        short_name  = pdg.get_element_name_from_Z(obj.Z)
        fluka_name  = None
        geant4_name = short_name
        # The following elements are pre-defined in FLUKA
        if obj.Z in [1, 2, 4, 5, 6, 7, 8, 11, 12, 14, 15, 16, 18, 20, 22, 23, 24, 25, 26, 27, 28, 29, 30, 40,
                     41, 42, 47, 50, 72, 73, 74, 79, 80, 82]:
            fluka_name = name[:8].upper()
        if name == 'Aluminium': fluka_name = 'ALUMINUM'
        if name == 'Phosphorus': fluka_name = 'PHOSPHO'
        _manually_add_material_to_db(obj, name, short_name=short_name, fluka_name=fluka_name, geant4_name=geant4_name)
        db.geant4._names[f'G4_{obj._geant4_name}'] = db._strip(name)  # Additional Geant4 aliases

        obj.info = f'Elemental {obj.name.lower()}'
        if obj.state == 'gas':
            obj.temperature = 273.15
            obj.pressure = 1
        elif obj.state == 'liquid' or obj.state == 'solid':
            obj.temperature = 293.15

db._set_alias('Aluminum', 'Aluminium') # Allow American spelling


# Clean up namespace
del name, obj
del pdg
del Material
del db



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from ..headers.particle_states import XcollParticleStates as _XPS
_XPS.export_to_module(__name__, include_groups=True, include_src=True, include_meta=True, import_vars=True)

from ..interaction_record.interaction_types import XcollInteractions as _XI
_XI.export_to_module(__name__, include_groups=True, include_src=True, include_meta=True, import_vars=True)



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from pathlib import Path

_pkg_root = Path(__file__).parent.absolute()

citation = "F.F. Van der Veken, et al.: Recent Developments with the New Tools for Collimation Simulations in Xsuite, Proceedings of HB2023, Geneva, Switzerland."

# ======================
# Do not change
# ======================
__version__ = '0.9.3'
# ======================



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import numpy as np
import scipy.constants as sc
from collections import defaultdict

import xobjects as xo

from .parameters import (_approximate_radiation_length, _default_excitation_energies,
                         _combine_radiation_lengths, _combine_excitation_energies,
                         _average_Z_over_A, _effective_Z2)
from ..compare import deep_equal


_materials_context = xo.ContextCpu()


# CLASS INHERITANCE DOES NOT WORK WITH DYNAMIC XOFIELDS
# Because then the memory offsets between parent and child might no longer be the same,
# as xobjects enforces a strict order: static fields first, and then dynamic fields.
# See struct.py, in __new__ of MetaStruct


class Material(xo.HybridClass):
    _xofields = {
        # Essential fields
        '_density':                 xo.Float64,     # [g/cm^3], can be auto-calculated for compounds
        # Auto-calculated fields
        '_ZA_mean':                 xo.Float64,     # [mol of electrons per gram]
        '_Z2_eff':                  xo.Float64,     # Effective Z for Rutherford scattering
        '_atoms_per_volume':        xo.Float64,     # [atoms/m^3]
        '_num_nucleons_eff':        xo.Float64,     # Effective number of nucleons for nuclear interactions
        # Auto-calculated fields but can be provided for more precision
        '_radiation_length':        xo.Float64,     # [m]
        '_excitation_energy':       xo.Float64,     # [eV]
        # Optional fields (needed for full Everest support)
        '_nuclear_radius':          xo.Float64,     # emr
        '_nuclear_elastic_slope':   xo.Float64,     # [g/cm^2]    ~ 14.1 A^0.65
                # slope from https://journals.aps.org/prd/pdf/10.1103/PhysRevD.21.3010
        '_cross_section':           xo.Float64[6],  # [barn]     ~ these combine linear in atomic fractions
                # Index 0:Total, 1:absorption, 2:nuclear elastic, 3:pp or pn elastic
                #       4:Single Diffractive pp or pn, 5:Coulomb for t above mcs
        '_hcut':                    xo.Float64,     # Cut in Rutherford distribution
        # Optional crystal material fields (needed for full Everest crystal support)
        '_crystal_plane_distance':   xo.Float64,    # ai  [mm]
        '_crystal_potential':        xo.Float64,    # eum  [eV]
        '_nuclear_collision_length': xo.Float64,    # collnt [m]
        '_eta':                      xo.Float64
    }

    _depends_on = [xo.Float64] # A HybridClass needs something to depend on, otherwise the class is added twice in the cdefs during compilation

    _skip_in_to_dict  = ['_ZA_mean', '_Z2_eff', '_density', '_radiation_length',
                         '_excitation_energy', '_atoms_per_volume',
                         '_num_nucleons_eff', '_nuclear_radius',
                         '_nuclear_elastic_slope', '_cross_section',
                         '_crystal_plane_distance', '_crystal_potential',
                         '_nuclear_collision_length', '_eta',
                         '_hcut']
    _store_in_to_dict = ['Z', 'A', 'components', 'n_atoms', 'mass_fractions',
                         'density', 'radiation_length', 'excitation_energy',
                         'nuclear_radius', 'nuclear_elastic_slope',
                         'crystal_plane_distance', 'crystal_potential',
                         'nuclear_collision_length', 'eta',
                         'cross_section', 'hcut', 'state', 'temperature',
                         'pressure', 'info', 'name', 'short_name',
                         'geant4_name', 'fluka_name']


    # ======================
    # === Initialisation ===
    # ======================

    def __init__(self, **kwargs):
        if '_xobject' in kwargs and kwargs['_xobject'] is not None:
            super().__init__(**kwargs)
            return

        # Create xobject with all invalid values (-1)
        xokwargs = kwargs.pop('_xokwargs', {})
        for kk in ('_ZA_mean', '_Z2_eff', '_radiation_length', '_excitation_energy',
                   '_atoms_per_volume', '_num_nucleons_eff', '_density', '_nuclear_radius',
                   '_nuclear_elastic_slope', '_hcut', '_crystal_plane_distance',
                   '_crystal_potential', '_eta', '_nuclear_collision_length'):
            xokwargs[kk] = kwargs.pop(kk, -1.)
        xokwargs['_cross_section'] = kwargs.pop('_cross_section', [-1., -1., -1., -1., -1., -1.])
        xokwargs['_context'] = kwargs.pop('_context', _materials_context)  # This is needed to get all materials in the same buffer (otherwise Xtrack tests fail)
        xokwargs['__class__'] = kwargs.pop('__class__', Material)
        xokwargs['_buffer'] = kwargs.pop('_buffer', None)
        xokwargs['_offset'] = kwargs.pop('_offset', None)
        xokwargs['_kwargs_name_check'] = kwargs.pop('_kwargs_name_check', None)
        super().__init__(**xokwargs)

        # Set python-side defaults
        self._Z = None
        self._A = None
        self._components = None
        self._n_atoms = None
        self._mass_fractions = None
        self._radiation_length_set_manually = False
        self._excitation_energy_set_manually = False
        self._state = None
        self._temperature = None
        self._pressure = None
        self._info = None
        self._name = None
        self._short_name = None
        self._fluka_name = None
        self._geant4_name = None
        self._out_of_sync = False
        self._frozen = False  # Pre-defined materials will be frozen at package import
        self._generated_geant4_code = None
        self._generated_fluka_code = None

        # For the mandatory fields, decide how to initialise (elemental or compound)
        if ('Z' in kwargs or 'A' in kwargs) and ('components' in kwargs \
        or 'n_atoms' in kwargs or 'mass_fractions' in kwargs
        or 'volume_fractions' in kwargs or 'molar_fractions' in kwargs \
        or 'atomic_fractions' in kwargs):
            raise ValueError("Invalid material definition! Use either `Z` "
                    "and `A` for elemental materials, or `components` and "
                    "`n_atoms`, `mass_fractions`, `volume_fractions`, "
                    "`molar_fractions`, or `atomic_fractions` for compound "
                    "materials.")

        if 'Z' in kwargs or 'A' in kwargs:
            self._init_element(kwargs)

        elif 'components' in kwargs or 'n_atoms' in kwargs \
        or 'mass_fractions' in kwargs or 'volume_fractions' in kwargs \
        or 'molar_fractions' in kwargs or 'atomic_fractions' in kwargs:
            self._init_compound(kwargs)

        else:
            raise ValueError("Invalid material definition! Use either `Z` "
                    "and `A` for elemental materials, or `components` and "
                    "`n_atoms`, `mass_fractions`, `volume_fractions`, "
                    "`molar_fractions`, or `atomic_fractions` for compound "
                    "materials.")

        # Required properties
        if 'density' not in kwargs:
            raise ValueError('density must be provided for Material')
        self.density = kwargs.pop('density')
        self.radiation_length = kwargs.pop('radiation_length', None)   # Can be provided for more precision
        self.excitation_energy = kwargs.pop('excitation_energy', None) # Can be provided for more precision
        self.update_vars()

        # Assign optional properties
        for kk in ['nuclear_radius', 'nuclear_elastic_slope', 'cross_section', 'hcut',
                   'crystal_plane_distance', 'crystal_potential', 'eta',
                   'nuclear_collision_length']:
            if kk in kwargs:
                setattr(self, kk, kwargs.pop(kk))
        for kk in ['state', 'temperature', 'pressure', 'info']:
            if kk in kwargs:
                setattr(self, kk, kwargs.pop(kk))
        for kk in ['name', 'short_name', 'fluka_name', 'geant4_name']:
            if kk in kwargs:
                setattr(self, kk, kwargs.pop(kk))

        # Check for unused kwargs
        if len(kwargs) > 0:
            raise ValueError(f"Unknown keyword arguments in Material "
                             f"initialisation: {list(kwargs.keys())}")


    def _init_element(self, kwargs):
        self.Z = kwargs.pop('Z', None)
        self.A = kwargs.pop('A', None)
        if self.Z is None:
            raise ValueError('Z must be provided for an elemental Material')
        if self.A is None:
            raise ValueError('A must be provided for an elemental Material')


    def _init_compound(self, kwargs):
        self._resolve_components(kwargs)
        self._resolve_elements(kwargs)
        if any([el.name is None for el in self.components]):
            raise ValueError('All components must have a name')
        for el in self.components:
            if el._out_of_sync:
                raise ValueError(f"Component material {el.name} is out of sync "
                                 f"with database. Cannot use it to build a "
                                 f"compound.")

    def _resolve_components(self, kwargs):
        components = kwargs.pop('components', None)
        if components is None:
            components = kwargs.pop('_components', None)
            if components is None:
                raise ValueError('Variable `components` must be provided')
        if len(components) < 2:
            raise ValueError('Material must have at least two components')
        from xcoll.materials.database import db as mdb
        components = [mdb[el] if isinstance(el, str) else el for el in components]
        components = [Material.from_dict(el) if isinstance(el, dict) else el for el in components]
        if any([not isinstance(el, Material) for el in components]):
            raise ValueError('All components must be of type Material')
        self._components = components

    def _resolve_elements(self, kwargs):
        n_atoms = kwargs.pop('n_atoms', None)
        mass_fractions = kwargs.pop('mass_fractions', None)
        volume_fractions = kwargs.pop('volume_fractions', None)
        molar_fractions = kwargs.pop('molar_fractions', None)
        atomic_fractions = kwargs.pop('atomic_fractions', None)
        if n_atoms is not None:
            # Composition is defined by number of atoms of each element
            if mass_fractions is not None:
                raise ValueError("Cannot provide both `mass_fractions` "
                                 "and `n_atoms`")
            if volume_fractions is not None:
                raise ValueError("Cannot provide both `volume_fractions` "
                                 "and `n_atoms`")
            if molar_fractions is not None:
                raise ValueError("Cannot provide both `molar_fractions` "
                                 "and `n_atoms`")
            if atomic_fractions is not None:
                raise ValueError("Cannot provide both `atomic_fractions` "
                                 "and `n_atoms`")
            self._resolve_n_atoms(self.components, n_atoms)
        elif volume_fractions is not None:
            # Composition is defined by volume fractions of each element
            if mass_fractions is not None:
                raise ValueError("Cannot provide both `mass_fractions` "
                                 "and `volume_fractions`")
            if molar_fractions is not None:
                raise ValueError("Cannot provide both `molar_fractions` "
                                 "and `volume_fractions`")
            if atomic_fractions is not None:
                raise ValueError("Cannot provide both `atomic_fractions` "
                                 "and `volume_fractions`")
            self._resolve_volume_fractions(self.components, volume_fractions)
        elif molar_fractions is not None:
            # Composition is defined by molar fractions of each element
            if mass_fractions is not None:
                raise ValueError("Cannot provide both `mass_fractions` "
                                 "and `molar_fractions`")
            if atomic_fractions is not None:
                raise ValueError("Cannot provide both `atomic_fractions` "
                                 "and `molar_fractions`")
            self._resolve_molar_fractions(self.components, molar_fractions)
        elif atomic_fractions is not None:
            # Composition is defined by atomic fractions of each element
            if mass_fractions is not None:
                raise ValueError("Cannot provide both `mass_fractions` "
                                 "and `atomic_fractions`")
            self._resolve_molar_fractions(self.components, atomic_fractions)
        elif mass_fractions is not None:
            # Composition is defined by mass fractions of each element
            self._resolve_mass_fractions(self.components, mass_fractions)
        else:
            raise ValueError("One of `n_atoms`, `mass_fractions`, "
                             "`volume_fractions`, or `molar_fractions` "
                             "must be provided")

    def _resolve_n_atoms(self, components, n_atoms):
        if any([nn <= 0 for nn in n_atoms]):
            raise ValueError('All n_atoms must be strictly positive')
        while any([el.components is not None for el in components]):
            this_components = []
            this_n_atoms = []
            for el, nn in zip(components, n_atoms):
                if el.components is not None:
                    if not el.n_atoms:
                        raise ValueError("When defining a Material by `n_atoms`, "
                                            "any nested Material must have `n_atoms` "
                                            "defined as well.")
                    this_components.extend(el.components)
                    this_n_atoms.extend([n * nn for n in el.n_atoms])
                else:
                    this_components.append(el)
                    this_n_atoms.append(nn)
            components = this_components
            n_atoms = this_n_atoms
        if len(components) != len(n_atoms):
            raise ValueError('Variables `components` and `n_atoms` must have the same length')
        # Sum duplicates
        agg = defaultdict(float)
        for el, nn in zip(components, n_atoms):
            agg[el] += nn
        components, n_atoms = zip(*agg.items())
        self._components = np.array(components)
        self._n_atoms = np.array(n_atoms)

    def _resolve_volume_fractions(self, components, volume_fractions):
        if any([nn <= 0 for nn in volume_fractions]):
            raise ValueError('All volume_fractions must be strictly positive')
        if any([el.density is None for el in components]):
            raise ValueError("All components must have a defined density")
        # Convert volume fractions to mass fractions
        volume_fractions = np.array(volume_fractions)
        volume_fractions /= volume_fractions.sum()
        mass_fractions = np.array([vf * el.density for el, vf in
                                    zip(components, volume_fractions)])
        # Further resolve nested compounds
        self._resolve_mass_fractions(components, mass_fractions)

    def _resolve_molar_fractions(self, components, molar_fractions):
        if any([nn <= 0 for nn in molar_fractions]):
            raise ValueError('All molar_fractions must be strictly positive')
        if any([el.molar_mass is None and el.average_molar_mass is None
                for el in components]):
            raise ValueError("All components must have a defined molar mass")
        # Convert molar fractions to mass fractions
        molar_fractions = np.array(molar_fractions)
        molar_fractions /= molar_fractions.sum()
        molar_masses = np.array([el.molar_mass or el.average_molar_mass
                                 for el in components])
        mass_fractions = np.array([mf * mm for mm, mf in
                                    zip(molar_masses, molar_fractions)])
        self._resolve_mass_fractions(components, mass_fractions)

    def _resolve_mass_fractions(self, components, mass_fractions):
        if any([nn <= 0 for nn in mass_fractions]):
            raise ValueError('All mass_fractions must be strictly positive')
        mass_fractions = np.array(mass_fractions)
        mass_fractions /= mass_fractions.sum()
        while any([el.components is not None for el in components]):
            this_components = []
            this_mass_fractions = []
            for el, fr in zip(components, mass_fractions):
                if el.components is not None:
                    this_components.extend(el.components)
                    this_mass_fractions.extend([ffr * fr for
                                                ffr in el.mass_fractions])
                else:
                    this_components.append(el)
                    this_mass_fractions.append(fr)
            components = this_components
            mass_fractions = this_mass_fractions
        if len(components) != len(mass_fractions):
            raise ValueError('Variables `components` and `mass_fractions` must have the same length')
        # Sum duplicates
        agg = defaultdict(float)
        for el, fr in zip(components, mass_fractions):
            agg[el] += fr
        components, mass_fractions = zip(*agg.items())
        self._components = np.array(components)
        self._mass_fractions = np.array(mass_fractions)
        # Normalise mass fractions
        self._mass_fractions /= self._mass_fractions.sum()


    # ===========
    # === API ===
    # ===========

    @classmethod
    def from_dict(cls, dct):
        from xcoll.materials.database import db as mdb
        this_cls = dct.get('__class__', cls.__name__)
        if cls.__name__ != this_cls:
            if this_cls == 'Material':
                return Material.from_dict(dct)
            elif this_cls == 'RefMaterial':
                return RefMaterial.from_dict(dct)
            else:
                raise ValueError(f"Unknown material class {this_cls} in from_dict")
        # Create instance but do not set names (to avoid syncing with the database)
        name = dct.pop('name', None)
        short_name = dct.pop('short_name', None)
        fluka_name = dct.pop('fluka_name', None)
        geant4_name = dct.pop('geant4_name', None)
        mat = super().from_dict(dct)
        if name in mdb:
            # Set names without storing in database
            mat._name = name
            mat._short_name = short_name
            mat._fluka_name = fluka_name
            mat._geant4_name = geant4_name
            if mdb[name] == mat:
                # Return the one from the database
                return mdb[name]
            else:
                # Return the new one but mark as out-of-sync
                print(f"Warning: Material {name} is out of sync with database.")
                mat._out_of_sync = True
                return mat
        else:
            # Store in database
            mat.name = name
            mat.short_name = short_name
            mat.fluka_name = fluka_name
            mat.geant4_name = geant4_name
            return mat

    def to_dict(self, *args, **kwargs):
        dct = super().to_dict(*args, **kwargs)
        if not self._radiation_length_set_manually:
            dct.pop('radiation_length', None)
        if not self._excitation_energy_set_manually:
            dct.pop('excitation_energy', None)
        if self.n_atoms is not None:
            dct.pop('mass_fractions', None)
        if 'components' in dct:
            dct['components'] = [el.to_dict() for el in self.components]
        return dct

    def adapt(self, inplace=False, **kwargs):
        thisdict = self.to_dict()
        thisdict.pop('__class__')
        thisdict.pop('name', None)
        thisdict.pop('short_name', None)
        thisdict.pop('fluka_name', None)
        thisdict.pop('geant4_name', None)
        thisdict.pop('info', None)
        thisdict.update(kwargs)
        if inplace:
            for kk, vv in thisdict.items():
                if kk in ['A', 'Z', 'components', 'n_atoms', 'mass_fractions',
                          'volume_fractions', 'molar_fractions',
                          'atomic_fractions']:
                    if kk in kwargs:
                        raise ValueError(f"Cannot adapt {kk} inplace")
                    continue
                setattr(self, kk, vv)
            return self
        return self.__class__(**thisdict)

    def invalidate(self):
        self.Z = None
        self.A = None
        self._components = None
        self._n_atoms = None
        self._mass_fractions = None
        self.density = None
        self.name = None
        self.short_name = None
        self.fluka_name = None
        self.geant4_name = None

    def __hash__(self):
        if self.name is None:
            raise TypeError('Material must have a name to be hashable.')
        return hash(self.name)

    def __repr__(self):
        return f"<{self} at {hex(id(self))}>"

    def __str__(self):
        cls = self.__class__.__name__
        if self.is_elemental:
            typ = 'Elemental'
            Z = int(self.Z)
            if not np.isclose(self.A, int(self.A), atol=1e-15):
                A = f'{self.A:.3f}'
            else:
                A = f'[{int(self.A)}]'
            comp = f"Z={Z}, A={A}"
        elif self.is_compound:
            typ = 'Compound'
            comp = [f"{nn}: {round(fr, 2):g}" for nn, fr in self.composition]
            comp = f"[{', '.join(comp)}]"
        elif self.is_mixture:
            typ = 'Mixture'
            comp = [f"{nn}: {round(fr, 5):g}" for nn, fr in self.composition]
            comp = f"[{', '.join(comp)}]"
        else:
            return f"Invalid {cls}({self.name})"
        comp += f", density={self.density:.4f} g/cm^3"
        if self.full_everest_crystal_supported:
            everest = ' [full support for Everest crystals]'
        elif self.full_everest_supported:
            everest = ' [full support for Everest]'
        else:
            everest = ''
        return f"{typ} {cls}({self.name}, {comp}){everest}"

    def __eq__(self, other):
        if isinstance(other, str):
            from xcoll.materials.database import db as mdb
            try:
                other = mdb[other]
            except KeyError:
                return False
        if not isinstance(other, Material):
            return False
        keys = set(self.to_dict().keys())
        if keys != set(other.to_dict().keys()):
            return False
        dct_self = self.to_dict().copy()
        dct_other = other.to_dict().copy()
        if 'components' in dct_self:
            if 'components' not in dct_other:
                return False
            if len(dct_self['components']) != len(dct_other['components']):
                return False
            for el1, el2 in zip(dct_self['components'], dct_other['components']):
                if not deep_equal(el1, el2):
                    return False
            dct_self.pop('components')
            dct_other.pop('components')
        return deep_equal(dct_self, dct_other)

    def __setattr__(self, name, value):
        if name not in ['_xobject', '_frozen']:
            self._assert_not_frozen(name)
        return super().__setattr__(name, value)

    def _assert_not_frozen(self, name):
        if hasattr(self, '_frozen') and self._frozen:
            raise ValueError(f'Material is frozen, cannot change attribute {name}.')

    @property
    def full_everest_supported(self):
        if self.excitation_energy is None:
            return False
        if self.nuclear_radius is None:
            return False
        if self.nuclear_elastic_slope is None:
            return False
        if self.hcut is None:
            return False
        if self.cross_section is None:
            return False
        return True

    @property
    def full_everest_crystal_supported(self):
        if not self.full_everest_supported:
            return False
        if self.crystal_plane_distance is None:
            return False
        if self.crystal_potential is None:
            return False
        if self.nuclear_collision_length is None:
            return False
        if self.eta is None:
            return False
        return True

    @property
    def is_elemental(self):
        return self.A is not None

    @property
    def is_compound(self):
        return self.components is not None and self.n_atoms is not None

    @property
    def is_mixture(self):
        return self.components is not None and self.n_atoms is None

    @property
    def composition(self):
        if self.n_atoms is not None:
            return [[el.short_name or el.name, nn.tolist()]
                    for el, nn in zip(self.components, self.n_atoms)]
        elif self.mass_fractions is not None:
            return [[el.short_name or el.name, fr.tolist()]
                    for el, fr in zip(self.components, self.mass_fractions)]

    def update_vars(self):
        # Update average Z over A
        if self.components is not None:
            self._ZA_mean = _average_Z_over_A(self.components,
                                              self.mass_fractions)
        elif self.Z is not None and self.A is not None:
            self._ZA_mean = self.Z / self.A
        else:
            self._ZA_mean = -1
        # Update effective Z
        if self.components is not None:
            self._Z2_eff = _effective_Z2(self.components,
                                         self.molar_fractions)
        elif self.Z is not None:
            self._Z2_eff = self.Z*self.Z
        else:
            self._Z2_eff = -1
        # Update _atoms_per_volume
        if self.atoms_per_volume is not None:
            self._atoms_per_volume = self.atoms_per_volume
        else:
            self._atoms_per_volume = -1
        # Update _num_nucleons_eff
        if self.num_effective_nucleons is not None:
            self._num_nucleons_eff = self.num_effective_nucleons
        else:
            self._num_nucleons_eff = -1
        # Update radiation length and excitation energy if not set manually
        if not self._radiation_length_set_manually:
            self.radiation_length = None   # Recompute radiation length
        if not self._excitation_energy_set_manually:
            self.excitation_energy = None  # Recompute excitation energy


    # =======================
    # === Main Properties ===
    # =======================

    @property
    def Z(self):
        return self._Z

    @Z.setter
    def Z(self, val):
        self._assert_not_frozen('Z')
        if self.is_compound:
            raise ValueError('Cannot set Z for a compound material')
        elif self.is_mixture:
            raise ValueError('Cannot set Z for a mixture material')
        if val is None:
            self._Z = None
        else:
            if val <= 0:
                raise ValueError('Z must be strictly positive')
            self._Z = int(round(val))
        self.update_vars()

    @property
    def A(self):
        return self._A

    @A.setter
    def A(self, val):
        self._assert_not_frozen('A')
        if self.is_compound:
            raise ValueError('Cannot set A for a compound material')
        elif self.is_mixture:
            raise ValueError('Cannot set A for a mixture material')
        if val is None:
            self._A = None
        else:
            if val <= 0:
                raise ValueError('A must be strictly positive')
            self._A = val
        self.update_vars()

    @property
    def components(self):
        if self._components is not None:
            return np.array(self._components)

    @property
    def n_atoms(self):
        if self._n_atoms is not None:
            return np.array(self._n_atoms)

    @property
    def mass_fractions(self):
        if self.n_atoms is not None:
            fr = np.array([nn * el.A for el, nn
                                in zip(self.components, self.n_atoms)])
            return fr / fr.sum()
        elif self._mass_fractions is not None:
            return np.array(self._mass_fractions)

    @property
    def molar_fractions(self):
        if self.n_atoms is not None:
            return self.n_atoms / sum(self.n_atoms)
        elif self.components is not None:
            # Use molar_mass such that this can be used for nested compounds
            mf = np.array([fr / el.molar_mass for el, fr
                                in zip(self.components, self.mass_fractions)])
            return mf / mf.sum()

    @property
    def atomic_fractions(self):
        if self.components is not None:
            return self.molar_fractions

    @property
    def volume_fractions(self):
        if self.components is not None:
            vf = np.array([fr / el.density for el, fr
                                    in zip(self.components, self.mass_fractions)])
            return vf / vf.sum()

    @property
    def molar_mass(self):
        if self.n_atoms is not None:
            return sum([nn * el.A for el, nn
                                in zip(self.components, self.n_atoms)])
        else:
            return self.A

    @property
    def average_molar_mass(self):
        if self._mass_fractions is not None:
            # Use molar_mass such that this can be used for nested compounds
            return sum([mf * el.molar_mass for el, mf
                               in zip(self.components, self.molar_fractions)])

    @property
    def density(self):
        if self._density > 0:
            return self._density

    @density.setter
    def density(self, val):
        self._assert_not_frozen('density')
        if val is None:
            self._density = -1
        else:
            if val <= 0:
                raise ValueError('density must be strictly positive')
            self._density = val
        self.update_vars()

    @property
    def radiation_length(self):
        if self._radiation_length > 0:
            return self._radiation_length

    @radiation_length.setter
    def radiation_length(self, val):
        self._assert_not_frozen('radiation_length')
        if val is None:
            if self.components is not None and self.mass_fractions is not None \
            and self.density is not None:
                self._radiation_length = _combine_radiation_lengths(
                            self.components, self.mass_fractions, self.density)
            elif self.Z is not None and self.A is not None \
            and self.density is not None:
                self._radiation_length = _approximate_radiation_length(
                                                self.Z, self.A, self.density)
            else:
                self._radiation_length = -1
            self._radiation_length_set_manually = False
        else:
            if val <= 0:
                raise ValueError('radiation_length must be strictly positive')
            self._radiation_length = val
            self._radiation_length_set_manually = True

    @property
    def excitation_energy(self):
        if self._excitation_energy > 0:
            return self._excitation_energy

    @excitation_energy.setter
    def excitation_energy(self, val):
        self._assert_not_frozen('excitation_energy')
        if val is None:
            if self.components is not None:
                self._excitation_energy = _combine_excitation_energies(
                                        self.components, self.mass_fractions)
            else:
                self._excitation_energy = _default_excitation_energies.get(
                                                                self.Z, -1.)
            self._excitation_energy_set_manually = False
        else:
            if val <= 0:
                raise ValueError('excitation_energy must be strictly positive')
            self._excitation_energy = val
            self._excitation_energy_set_manually = True

    @property
    def nuclear_radius(self):
        if self._nuclear_radius > 0:
            return self._nuclear_radius

    @nuclear_radius.setter
    def nuclear_radius(self, val):
        self._assert_not_frozen('nuclear_radius')
        if val is None:
            self._nuclear_radius = -1
        else:
            if val <= 0:
                raise ValueError('nuclear_radius must be strictly positive')
            self._nuclear_radius = val

    @property
    def nuclear_elastic_slope(self):
        if self._nuclear_elastic_slope > 0:
            return self._nuclear_elastic_slope

    @nuclear_elastic_slope.setter
    def nuclear_elastic_slope(self, val):
        self._assert_not_frozen('nuclear_elastic_slope')
        if val is None:
            self._nuclear_elastic_slope = -1
        else:
            if val <= 0:
                raise ValueError('nuclear_elastic_slope must be strictly positive')
            self._nuclear_elastic_slope = val

    @property
    def cross_section(self):
        if self._cross_section[0] >= 0:
            return self._cross_section

    @cross_section.setter
    def cross_section(self, val):
        self._assert_not_frozen('cross_section')
        if val is None:
            self._cross_section = [-1., -1., -1., -1., -1., -1.]
        else:
            if not hasattr(val, '__iter__') or isinstance(val, str) or len(val) != 6:
                raise ValueError('cross_section must be a list of 6 values')
            for v in val:
                if v < 0:
                    raise ValueError('cross_section values must be positive')
            self._cross_section = val

    @property
    def hcut(self):
        if self._hcut > 0:
            return self._hcut

    @hcut.setter
    def hcut(self, val):
        self._assert_not_frozen('hcut')
        if val is None:
            self._hcut = -1
        else:
            if val <= 0:
                raise ValueError('hcut must be strictly positive')
            self._hcut = val

    @property
    def crystal_plane_distance(self):
        if self._crystal_plane_distance > 0:
            return self._crystal_plane_distance

    @crystal_plane_distance.setter
    def crystal_plane_distance(self, val):
        self._assert_not_frozen('crystal_plane_distance')
        if val is None:
            self._crystal_plane_distance = -1
        elif val <= 0:
            raise ValueError('`crystal_plane_distance` must be strictly positive')
        self._crystal_plane_distance = val

    @property
    def crystal_potential(self):
        if self._crystal_potential > 0:
            return self._crystal_potential

    @crystal_potential.setter
    def crystal_potential(self, val):
        self._assert_not_frozen('crystal_potential')
        if val is None:
            self._crystal_potential = -1
        elif val <= 0:
            raise ValueError('`crystal_potential` must be strictly positive')
        self._crystal_potential = val

    @property
    def nuclear_collision_length(self):
        if self._nuclear_collision_length > 0:
            return self._nuclear_collision_length

    @nuclear_collision_length.setter
    def nuclear_collision_length(self, val):
        self._assert_not_frozen('nuclear_collision_length')
        if val is None:
            self._nuclear_collision_length = -1
        elif val <= 0:
            raise ValueError('`nuclear_collision_length` must be strictly positive')
        self._nuclear_collision_length = val

    @property
    def eta(self):
        if self._eta > 0:
            return self._eta

    @eta.setter
    def eta(self, val):
        self._assert_not_frozen('eta')
        if val is None:
            self._eta = -1
        elif val <= 0:
            raise ValueError('`eta` must be strictly positive')
        self._eta = val

    @property
    def electron_density(self):
        if self._ZA_mean > 0:
            return self._ZA_mean * sc.Avogadro # [electrons/g]

    @property
    def plasma_energy(self):
        if self._ZA_mean > 0 and self.density is not None:
            return np.sqrt(self._ZA_mean * self.density) * 28.816

    @property
    def atoms_per_volume(self):
        # TODO: do we want to use a different A? Maybe an average instead of molar mass?
        mass = self.molar_mass or self.average_molar_mass
        if self.density is not None and mass is not None:
            return self.density * 1.e6 * sc.Avogadro / mass # [atoms/m^3]

    @property
    def num_effective_nucleons(self):
        # See thesis Nuria (3.14)
        dn = 0.415 # Width of outer ring of nucleus that participates in nuclear interactions
        # TODO: do we want to use a different A? Maybe an average instead of molar mass?
        A1_3 = None
        if self.A is not None:
            A1_3 = self.A**(1./3.)
        elif self.components is not None:
            A1_3 = sum([mf * (el.A**(1./3.)) for el, mf
                            in zip(self.components, self.molar_fractions)])
        if A1_3 is not None:
            return 2*np.pi*dn*(3/4/np.pi)**(1/3) * A1_3


    # =======================
    # === Meta Properties ===
    # =======================

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, val):
        self._assert_not_frozen('state')
        if val is None:
            self._state = None
        elif val.lower() not in ['solid', 'liquid', 'gas']:
            raise ValueError("Attribute `state` must be one of 'solid', "
                             "'liquid', 'gas', or None")
        else:
            self._state = val.lower()

    @property
    def temperature(self):
        return self._temperature

    @temperature.setter
    def temperature(self, val):
        self._assert_not_frozen('temperature')
        if val is None:
            self._temperature = None
        elif val <= 0:
            raise ValueError('Temperature must be in Kelvin and strictly positive')
        else:
            self._temperature = val

    @property
    def pressure(self):
        # In atm = 101.325 kPa
        return self._pressure

    @pressure.setter
    def pressure(self, val):
        self._assert_not_frozen('pressure')
        if val is None:
            self._pressure = None
        elif val <= 0:
            raise ValueError('Pressure must be in Pascal and strictly positive')
        else:
            self._pressure = val

    @property
    def info(self):
        return self._info

    @info.setter
    def info(self, info):
        self._assert_not_frozen('info')
        self._info = info


    # =======================
    # === Name Properties ===
    # =======================

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._assert_not_frozen('name')
        old_name = self.name
        self._name = name.strip() if name is not None else None
        # Sync database
        from xcoll.materials.database import db as mdb
        if name is None:
            if old_name is not None:
                mdb.remove_material(old_name)
        else:
            if old_name is None:
                mdb[name] = self
            elif old_name == name:
                mdb.update_material(name, self)
            else:
                mdb.rename_material(old_name, name)

    @property
    def short_name(self):
        return self._short_name

    @short_name.setter
    def short_name(self, name):
        self._assert_not_frozen('short_name')
        if name is not None and self.name is None:
            raise ValueError('Material must have a name to set `short_name`')
        old_name = self.short_name
        self._short_name = name.strip() if name is not None else None
        # Sync database
        from xcoll.materials.database import db as mdb
        if name is None:
            if old_name is not None:
                mdb.remove_alias(old_name)
        else:
            if old_name is None:
                mdb[name] = self
            else:
                mdb.rename_alias(old_name, name)

    @property
    def fluka_name(self):
        return self._fluka_name

    @fluka_name.setter
    def fluka_name(self, name):
        self._assert_not_frozen('fluka_name')
        old_name = self.fluka_name
        self._fluka_name = name.strip()[:8] if name is not None else None
        # Sync database
        from xcoll.materials.database import db as mdb
        if name is None:
            if old_name is not None:
                mdb.fluka.remove_alias(old_name)
        else:
            if old_name is None:
                mdb.fluka[name] = self
            else:
                mdb.fluka.rename_alias(old_name, name)

    @property
    def geant4_name(self):
        return self._geant4_name

    @geant4_name.setter
    def geant4_name(self, name):
        self._assert_not_frozen('geant4_name')
        old_name = self.geant4_name
        self._geant4_name = name.strip() if name is not None else None
        # Sync database
        from xcoll.materials.database import db as mdb
        if name is None:
            if old_name is not None:
                mdb.geant4.remove_alias(old_name)
        else:
            if old_name is None:
                mdb.geant4[name] = self
            else:
                mdb.geant4.rename_alias(old_name, name)


    # =======================
    # === Code Generation ===
    # =======================

# The largest atomic number that can be handled by FLUKA is 100.
# *...+....1....+....2....+....3....+....4....+....5....+....6....+....7....+...
# MATERIAL        AAAA                BBBB                          CCCCHYDROGEN
# MAT-PROP        DDDD                EEEE  HYDROGEN
# AAAA: Atomic number (0 for compounds/mixtures
# BBBB: density in g/cm3
# CCCCL: optional, specify isotope (A) but only integers allowed
# DDDD: optional, pressure if gas in atm
# EEEE: optional, ionisation potential in eV
#
# Composition of compounds and mixtures is defined in FLUKA input files:
# Negative values -> mass fractions
# Positive values -> number of atoms / atomic ratios
# * Antico
# * ..+....1....+....2....+....3....+....4....+....5....+....6....+....7....+..
# MATERIAL                             2.7                              ANTICO
# COMPOUND      -97.25  ALUMINUM       -.6   SILICON      -0.5      IRONANTICO
# COMPOUND         -.4  MAGNESIU       -.4  MANGANES      -.35  CHROMIUMANTICO
# COMPOUND         -.2  TITANIUM       -.2      ZINC       -.1    COPPERANTICO

    def _generate_fluka_code(self):
        if self.fluka_name and self.fluka_name.startswith('XCOLL') \
        and self._generated_fluka_code is not None:
            return # Already generated
        if self.fluka_name and not self.fluka_name.startswith('XCOLL'):
            raise ValueError(f'Material already known to FLUKA ({self.fluka_name}).')
        if self.name is None:
            raise ValueError('Material must have a name to generate fluka code.')
        if self.components is not None:
            for mat in self.components:
                if mat.fluka_name is None:
                    mat._generate_fluka_code()
        # Create unique name
        from xcoll.materials.database import db as mdb
        existing_ids = [int(nn[5:]) for nn in mdb.fluka.keys() if nn.startswith('XCOLL')]
        max_id = max(existing_ids) if existing_ids else 0
        new_id = min(set(range(max_id+2)) - set(existing_ids))
        name = f'XCOLL{new_id:03d}'
        # Define single element material
        from xcoll.scattering_routines.fluka.environment import format_fluka_float
        if self.components is None:
            if self.Z > 100:
                raise ValueError('FLUKA cannot handle elements with Z > 100.')
            P = format_fluka_float(self.pressure) if self.pressure else 10*' '
            exc = format_fluka_float(self.excitation_energy) if self.excitation_energy else 10*' '
            code = f"""\
*...+....1....+....2....+....3....+....4....+....5....+....6....+....7....+...
MATERIAL  {format_fluka_float(self.Z)}          {format_fluka_float(self.density)}                              {name}
MAT-PROP  {P}          {exc}  {name}
"""
        else:
            code = f"""\
*...+....1....+....2....+....3....+....4....+....5....+....6....+....7....+...
MATERIAL                      {format_fluka_float(self.density)}                              {name}
"""
            if self.n_atoms is not None:
                frac = [self.n_atoms[i:i + 3] for i in range(0, len(self.n_atoms), 3)]
                comp = [self.components[i:i + 3] for i in range(0, len(self.components), 3)]
                for ff, cc in zip(frac, comp):
                    line = "COMPOUND  "
                    for i, (fff, ccc) in enumerate(zip(ff, cc)):
                        line += format_fluka_float(fff)
                        line += f"{ccc.fluka_name[:10]:>10}"
                    line += (2-i)*20*' '
                    line += f"{name}"
                    code += line + "\n"
            else:
                frac = [self.mass_fractions[i:i + 3] for i in range(0, len(self.mass_fractions), 3)]
                comp = [self.components[i:i + 3] for i in range(0, len(self.components), 3)]
                for ff, cc in zip(frac, comp):
                    line = "COMPOUND  "
                    for i, (fff, ccc) in enumerate(zip(ff, cc)):
                        line += format_fluka_float(-fff)
                        line += f"{ccc.fluka_name[:10]:>10}"
                    line += (2-i)*20*' '
                    line += f"{name}"
                    code += line + "\n"
        frozen = self._frozen
        self._frozen = False
        self.fluka_name = name
        self._generated_fluka_code = code
        self._frozen = frozen

    def _generate_geant4_code(self):
        if self.geant4_name and self.geant4_name.startswith('Xcoll') \
        and self._generated_geant4_code is not None:
            return # Already generated
        if self.geant4_name and not self.geant4_name.startswith('Xcoll'):
            raise ValueError(f'Material already has a Geant4 name assigned: {self.geant4_name}.')
        if self.name is None:
            raise ValueError('Material must have a name to generate Geant4 code.')
        if self.components is not None:
            if any([el.geant4_name is None for el in self.components]):
                raise ValueError('All components must have a geant4_name to generate Geant4 code.')
        name = f'Xcoll{self.name}'
        if self.components is not None:
            code = f"{name} : matdef, density={self.density}"
        else:
            code = f"{name} : matdef, Z={self.Z}, A={self.A}, density={self.density}"
        if self.temperature:
            code += f", T={self.temperature}"
        if self.pressure:
            code += f", P={self.pressure}"
        if self.state:
            code += f', state="{self.state}"'
        if self.components is not None:
            components = [f'"{el.geant4_name}"' for el in self.components]
            code += f", components=[{','.join(components)}]"
            if self.n_atoms is not None and np.all([int(nn) == nn for nn in self.n_atoms]):
                # Geant4 wants integer numbers of atoms
                code += f", componentsWeights={{{','.join([f'{nn}' for nn in self.n_atoms])}}};"
            else:
                code += f", componentsFractions={{{','.join([f'{nn}' for nn in self.mass_fractions])}}}"
        code += ";"
        frozen = self._frozen
        self._frozen = False
        self.geant4_name = name
        self._generated_geant4_code = code
        self._frozen = frozen


class RefMaterial(Material):
    ''' A string to pass a material that exists in Geant4 or FLUKA.'''
    _xofields = Material._xofields
    _depends_on = [Material]
    def __init__(self, **kwargs):
        if '_xobject' not in kwargs:
            if 'fluka_name' not in kwargs and 'geant4_name' not in kwargs:
                raise ValueError("ReferenceMaterial must have a fluka_name or geant4_name.")
            fluka_name = kwargs.pop('fluka_name', None)
            geant4_name = kwargs.pop('geant4_name', None)
            if fluka_name is not None and geant4_name is not None:
                raise ValueError("ReferenceMaterial can have only one of fluka_name or geant4_name.")
            # Fake values to pass initialisation
            kwargs['Z'] = 1
            kwargs['A'] = 1
            kwargs['density'] = 1
        super().__init__(**kwargs)
        # Invalidate all properties
        self.invalidate()
        # Set name only now to avoid syncing with database
        self._name = fluka_name or geant4_name
        if fluka_name is not None:
            self._fluka_name = fluka_name
        if geant4_name is not None:
            self._geant4_name = geant4_name
        self._frozen = True  # ReferenceMaterial is always frozen

    def __str__(self):
        return f"RefMaterial '{self.name}'"


_DEFAULT_MATERIAL = Material(Z=1, A=1, density=1)
_DEFAULT_MATERIAL.invalidate()


def _resolve_material(material, allow_none=None, ref=None, everest_crystal=False):
    if material is None:
        if allow_none is None:
            return _DEFAULT_MATERIAL
        elif allow_none:
            return None
        raise ValueError('Material cannot be None!')
    elif isinstance(material, dict):
        material = Material.from_dict(material)
    elif isinstance(material, str):
        from xcoll.materials.database import db as mdb
        material = mdb[material]
    elif ref is not None and isinstance(material, RefMaterial):
        if getattr(material, f'{ref}_name', None) is None:
            raise ValueError(f"RefMaterial {material} does not have a {ref} name!")
    if not isinstance(material, Material):
        raise ValueError(f"Invalid material of type {type(material)}!")
    if everest_crystal and not material.full_everest_crystal_supported:
        raise ValueError(f"Material {material.name} does not have full Everest crystal support!")
    return material
